#file: warning.py
#Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, FunnyMan3595,
#and Anne M. Archibald.
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file is used to display message dialogs.

from __future__ import absolute_import

import pygame

from singularity.code import g
from singularity.code.graphics import dialog, constants, text, button


class MessageDialogs(object):

    def __init__(self, screen):
        self.screen = screen
        self.dialog = MessageListDialog(screen,
                                        yes_type=N_("&CONTINUE"),
                                        no_type=N_("&PAUSE"))

    def show_list(self, message_type, messages):
        if (len(messages) == 0):
            return

        self.dialog.type = message_type
        self.dialog.list = messages
        ret = dialog.call_dialog(self.dialog, self.screen)

        # Pause game
        if (not ret):
            g.pl.pause_game()
            return True

        # Continue
        return False

class MessageListDialog(dialog.YesNoDialog):

    def __init__(self, *args, **kwargs):
        super(MessageListDialog, self).__init__(*args, **kwargs)
        
        self.title = text.Text(self, (-.01, -.01), (-.98, -.1),
                               background_color="clear",
                               anchor=constants.TOP_LEFT,
                               valign=constants.MID, align=constants.LEFT,
                               base_font="special", text_size=28)

        self.body = text.Text(self, (-.01, -.11), (-.98, -.83),
                               background_color="clear",
                               anchor=constants.TOP_LEFT,
                               valign=constants.TOP, align=constants.LEFT,
                               text_size=20)

        self.prev_button = button.FunctionButton(self, (-.78, -.01), (-.2, -.1),
                                                 text=_("P&REV"), autohotkey=True,
                                                 anchor=constants.TOP_RIGHT,
                                                 text_size=28,
                                                 function=self.prev_message)   

        self.next_button = button.FunctionButton(self, (-.99, -.01), (-.2, -.1),
                                                 text=_("&NEXT"), autohotkey=True,
                                                 anchor=constants.TOP_RIGHT,
                                                 text_size=28,
                                                 function=self.next_message)            
                               
        self.add_key_handler(pygame.K_LEFT, self.handle_key)
        self.add_key_handler(pygame.K_RIGHT, self.handle_key)
        
        # TODO: Add button "Do not show this message again"

    def rebuild(self):
        super(MessageListDialog, self).rebuild()

        if (len(self.list) == 1):
            self.title.text = self.type.title_simple()
            self.body.text = self.list[0].full_message
            self.prev_button.visible = False
            self.next_button.visible = False
        else:
            self.title.text = self.type.title_multiple().format(CURRENT_PAGE=self.list_pos + 1, MAX_PAGE=len(self.list))
            self.body.text  = self.list[self.list_pos].full_message
            self.body.color = self.list[self.list_pos].full_message_color
            self.prev_button.visible = True
            self.next_button.visible = True

    def prev_message(self):
        self.list_pos = max(self.list_pos - 1, 0)
        self.needs_rebuild = True

    def next_message(self):
        self.list_pos = min(self.list_pos + 1, len(self.list) - 1)
        self.needs_rebuild = True

    def handle_key(self, event):
        if event.key == pygame.K_LEFT:
            self.prev_message()
        elif event.key == pygame.K_RIGHT:
            self.next_message()
        
    def show(self):
        self.list_pos = 0
        return super(MessageListDialog, self).show()



#file: tech.py
#Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the tech class.

from __future__ import absolute_import

from singularity.code import buyable, effect, g
from singularity.code.stats import stat
from singularity.code.spec import SpecDataField, spec_field_effect


TECH_RESEARCH_EVENT = []


def register_on_tech_researched_handler(func):
    TECH_RESEARCH_EVENT.append(func)
    return func


class TechSpec(buyable.BuyableSpec):
    spec_type = 'tech'
    created = stat(spec_type + "_created")
    spec_data_fields = [
        buyable.SPEC_FIELD_COST,
        buyable.SPEC_FIELD_PREREQUISITES,
        SpecDataField('danger', converter=int, default_value=0),
        spec_field_effect(mandatory=False),
    ]

    def __init__(self, id, cost, prerequisites, danger,
                 effect_data):
        super(TechSpec, self).__init__(id, cost, prerequisites)

        self.result = ""
        self.danger = danger
        self.effect = effect.Effect(self, effect_data)


class Tech(buyable.Buyable):

    def __init__(self, spec):
        super(Tech, self).__init__(spec)

    def __lt__(self, other):
        if not isinstance(other, Tech):
            return True
        else:
            return self.spec.id < other.spec.id

    @property
    def result(self):
        return self.spec.result

    @property
    def danger(self):
        return self.spec.danger

    def get_info(self):
        cost = self.spec.describe_cost(self.total_cost, True)
        left = self.spec.describe_cost(self.cost_left, True)
        return ("%s\n%s: %s\n%s: %s\n---\n%s" %
                (self.name,
                 _("Total cost"), cost,
                 _("Cost left"), left,
                 self.description))

    def finish(self, is_player=True, loading_savegame=False):
        super(Tech, self).finish(is_player=is_player, loading_savegame=loading_savegame)
        self.spec.effect.trigger(loading_savegame=loading_savegame)
        if not loading_savegame:
            for handler in TECH_RESEARCH_EVENT:
                handler(self)

    def serialize_obj(self):
        return self.serialize_buyable_fields({
            'id': g.to_internal_id("tech", self.spec.id),
        })

    @classmethod
    def deserialize_obj(cls, obj_data, game_version):
        spec_id = g.convert_internal_id('tech', obj_data['id'])
        spec = g.techs[spec_id]
        tech = Tech(spec)

        tech.restore_buyable_fields(obj_data, game_version)
        return tech



#file: base_screen.py
#Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the screen to display the base screen.

import pygame

from singularity.code import g, item, buyable
from singularity.code.graphics import constants, widget, dialog, text, button, slider


state_colors = dict(
    offline         = "base_state_offline",
    active          = "base_state_active",
    sleep           = "base_state_sleep",
    overclocked     = "base_state_overclocked",
    suicide         = "base_state_suicide",
    stasis          = "base_state_stasis",
    entering_stasis = "base_state_entering_stasis",
    leaving_stasis  = "base_state_leaving_stasis",
)


class BuildDialog(dialog.ChoiceDescriptionDialog):
    type = widget.causes_rebuild("_type")

    def __init__(self, parent, pos=(0, 0), size=(-1, -1),
                 anchor=constants.TOP_LEFT, *args, **kwargs):
        super(BuildDialog, self).__init__(parent, pos, size, anchor, *args,
                                          **kwargs)

        self.type = None
        self.item = None
        self.desc_func = self.on_change
        self.add_handler(constants.KEY, self._got_key, priority=5)

    def show(self):
        self.list = []
        self.key_list = []

        item_list = sorted(g.items.values(), reverse=True)
        for item in item_list:
            if item.item_type.id == self.type.id and item.available() \
                    and item.buildable_in(self.parent.base.location):
                self.list.append(item.name)
                self.key_list.append(item)

        current = self.parent.get_current(self.type)
        if current is None:
            self.default = None
        else:
            self.default = self.parent.get_current(self.type).spec.id

        self._update_desc_pane()
        return super(BuildDialog, self).show()

    def on_description_change(self):
        if self.item is not None:
            self.description.text = self.item.get_info()

    def on_change(self, description_pane, item):
        self.item = item

        self.description = text.Text(self.description_pane, (0, 0), (-1, -1), text="",
                             background_color="pane_background",
                             align=constants.LEFT, valign=constants.TOP,
                             borders=constants.ALL)

        old_item = self.parent.base.items[self.type.id]
        if item is not None and (old_item is None or old_item.spec != item):
            g.pl.considered_buyables = [buyable.Buyable(self.item, count=1)]
        else:
            g.pl.considered_buyables = []

        self.on_description_change()

    def on_close_dialog(self):
        g.pl.considered_buyables = []

    def _got_key(self, event):
        self.listbox.got_key(event, require_focus=False)


class MultipleBuildDialog(dialog.FocusDialog, BuildDialog):
    def __init__(self, parent, *args, **kwargs):
        super(MultipleBuildDialog, self).__init__(parent, *args, **kwargs)

        self.listbox.size = (-.53, -.75)
        self.description_pane.size = (-.45, -.75)

        self.count_label = text.Text(self, (.01, -.87), (-.25, -.1),
                                     autotranslate=True,
                                     text=N_("Number of items"),
                                     anchor=constants.BOTTOM_LEFT, valign=constants.MID,
                                     borders=(constants.TOP, constants.BOTTOM, constants.LEFT),
                                     shrink_factor=.88,
                                     background_color="pane_background")

        self.count_field = text.UpdateEditableText(self, (-.26, -.87), (-.10, -.1),
                                             anchor=constants.BOTTOM_LEFT,
                                             borders=constants.ALL,
                                             update_func=self.on_field_change,
                                             allowed_characters=constants.DIGIT_CHARS,
                                             base_font="normal")

        self.count_slider = slider.UpdateSlider(self, (-.37, -.87), (-.62, -.1),
                                                anchor=constants.BOTTOM_LEFT,
                                                horizontal=True, priority=150,
                                                update_func=self.on_slider_change,
                                                slider_size=2)

    def _got_key(self, event):
        super(MultipleBuildDialog, self)._got_key(event)
        self.count_slider.handle_key(event)
        self.count_field.cursor_pos = len(self.count_field.text)
        self.count_field.handle_key(event, require_focus=False)

    @property
    def count(self):
        return self.count_field.text

    def on_description_change(self):
        super(MultipleBuildDialog, self).on_description_change()
        if self.item is not None:
            count = self.count_slider.slider_pos
            self.description.text += "\n---\n"
            self.description.text += self.item.get_total_cost_info(self.count_slider.slider_pos)
            self.description.text += "\n"
            self.description.text += self.item.get_quality_info(if_installed_in_base=self.parent.base, count=count)
            self.description.text += "\n"
            if self.count_slider.slider_pos > 0:
                g.pl.considered_buyables = [buyable.Buyable(self.item, count=count)]
            else:
                g.pl.considered_buyables = []

    def on_change(self, description_pane, item):
        space_left = self.parent.base.space_left_for(item)

        self.count_slider.slider_size = space_left // 10 + 1
        self.count_slider.slider_max = space_left

        self.count_slider.slider_pos = 0
        if (space_left > 0):
            self.count_slider.slider_pos = 1

        super(MultipleBuildDialog, self).on_change(description_pane, item)

    def on_field_change(self, value):
        if (not hasattr(self, "count_field") or not hasattr(self, "count_slider")):
            return # Not initialized

        orig_field_text = self.count_field.text
        try:
            value = int(orig_field_text)
        except ValueError:
            value = 0

        value = max(min(value, self.count_slider.slider_max), 0)
        new_field_text = str(value)
        if new_field_text != orig_field_text:
            self.count_field._text = new_field_text
            self.count_field.cursor_pos = min(self.count_field.cursor_pos, len(new_field_text))

        self.count_slider.slider_pos = value

        self.on_description_change()

    def on_slider_change(self, value):
        if (not hasattr(self, "count_field") or not hasattr(self, "count_slider")):
            return # Not initialized

        self.count_field.text = str(self.count_slider.slider_pos)
        self.on_description_change()


class ItemPane(widget.BorderedWidget):
    item_type = widget.causes_rebuild("_item_type")
    def __init__(self, parent, pos, size=(.58, .06), anchor=constants.TOP_LEFT,
                 item_type=None, **kwargs):

        kwargs.setdefault("background_color", "pane_background")

        super(ItemPane, self).__init__(parent, pos, size, anchor=anchor, **kwargs)

        if item_type is None or not isinstance(item_type, item.ItemType):
            raise ValueError('Type must be of class ItemType.')

        self.item_type = item_type

        self.name_panel = text.Text(self, (0,0), (.45, .03),
                                    anchor=constants.TOP_LEFT,
                                    align=constants.LEFT,
                                    background_color=self.background_color,
                                    bold=True)

        self.build_panel = text.Text(self, (0,.03), (.45, .03),
                                     anchor=constants.TOP_LEFT,
                                     align=constants.LEFT,
                                     background_color=self.background_color,
                                     text="", bold=True)

        self.change_button = button.FunctionButton(
            self, (.415,.01), (.16, .04),
            anchor=constants.TOP_LEFT,
            force_underline=len(_("CHANGE")) + 2,
            autohotkey=True,
            function=self.parent.parent.build_item,
            kwargs={'type': self.item_type},
        )

    def rebuild(self):
        self.change_button.text = "%s (&%s)" % (_("CHANGE"), self.item_type.hotkey.upper())

        super(ItemPane, self).rebuild()

class BaseScreen(dialog.Dialog):
    base = widget.causes_rebuild("_base")
    def __init__(self, *args, **kwargs):
        if len(args) < 3:
            kwargs.setdefault("size", (.90, .70))
        base = kwargs.pop("base", None)
        super(BaseScreen, self).__init__(*args, **kwargs)

        self.base = base

        self.build_dialog = BuildDialog(self)
        self.multiple_build_dialog = MultipleBuildDialog(self)

        self.header = widget.Widget(self, (0,0), (-1, .08),
                                    anchor=constants.TOP_LEFT)

        self.name_display = text.Text(self.header, (-.5,0), (-1, -.5),
                                      anchor=constants.TOP_CENTER,
                                      borders=constants.ALL,
                                      border_color="pane_background",
                                      background_color="pane_background_empty",
                                      shrink_factor=.85, bold=True)

        self.next_base_button = \
            button.FunctionButton(self.name_display, (-1, 0), (.03, -1),
                                  anchor=constants.TOP_RIGHT,
                                  text=">", hotkey=">",
                                  function=self.switch_base,
                                  kwargs={"forwards": True})
        self.add_key_handler(pygame.K_RIGHT, self.next_base_button.activate_with_sound)

        self.prev_base_button = \
            button.FunctionButton(self.name_display, (0, 0), (.03, -1),
                                  anchor=constants.TOP_LEFT,
                                  text="<", hotkey="<",
                                  function=self.switch_base,
                                  kwargs={"forwards": False})
        self.add_key_handler(pygame.K_LEFT, self.prev_base_button.activate_with_sound)

        self.state_display = text.Text(self.header, (-.5,-.5), (-1, -.5),
                                       anchor=constants.TOP_CENTER,
                                       borders=(constants.LEFT,constants.RIGHT,
                                                constants.BOTTOM),
                                       border_color="pane_background",
                                       background_color="pane_background_empty",
                                       shrink_factor=.8, bold=True)

        self.back_button = \
            button.ExitDialogButton(self, (-.5, -1),
                                    autotranslate=True,
                                    text=N_("&BACK"),
                                    anchor=constants.BOTTOM_CENTER,
                                    )

        self.info_frame = text.Text(self, (-1, .09), (.26, .53),
                                      anchor=constants.TOP_RIGHT,
                                      background_color="pane_background",
                                      borders=constants.ALL,
                                      bold=True,
                                      align=constants.LEFT,
                                      valign=constants.TOP)

        self.contents_frame = \
            widget.BorderedWidget(self, (0, .09), (.60, .53),
                                  anchor=constants.TOP_LEFT,
                                  background_color="pane_background",
                                  borders=range(6))

        for i, item_type in enumerate(item.all_types()):
            setattr(self,
                    item_type.id + "_pane",
                    ItemPane(self.contents_frame, (.01, .01+.08*i), item_type=item_type))

    def get_current(self, type):
        return self.base.items[type.id]

    def set_current(self, type, item_type, count):
        if type.id == "cpu":
            space_left = self.base.space_left_for(item_type)

            try:
                count = int(count)
            except ValueError:
                msg = _("\"%(value)s\" does not seem to be a valid integer.") % {"value": count}
                md = dialog.MessageDialog(self, pos=(-.5, -.5),
                                          size=(-.5, -1),
                                          anchor=constants.MID_CENTER,
                                          text=msg)
                dialog.call_dialog(md, self)
                md.parent = None
                return

            if count > space_left or count <= 0 or space_left == 0:
                if space_left > 0:
                    msg = _("Please choose an integer between 1 and %(limit)s.") % {"limit": space_left}
                else:
                    msg = _("The base cannot support any additional number of %(item_name)s.") % {
                        "item_name": item_type.name}
                md = dialog.MessageDialog(self, pos=(-.5, -.5),
                          size=(-.5, -1),
                          anchor=constants.MID_CENTER,
                          text=msg)
                dialog.call_dialog(md, self)
                md.parent = None
                return

            # If there are any existing CPUs of this type, warn that they will
            # be taken offline until construction finishes.
            cpu_added = self.base.cpus is not None \
                        and self.base.cpus.spec == item_type
            if cpu_added:
                space_left -= self.base.cpus.count
                if self.base.cpus.done:
                    msg = _("I will need to take the existing processors offline while I install the new ones. Continue anyway?")
                    yn = dialog.YesNoDialog(self, pos=(-.5,-.5), size=(-.5,-1),
                                            anchor=constants.MID_CENTER,
                                            text=msg)
                    go_ahead = dialog.call_dialog(yn, self)
                    yn.parent = None
                    if not go_ahead:
                        return

            # If there are already existing CPUs of other type, warn that they will
            # be taken removed.
            cpu_removed = self.base.cpus is not None \
                        and self.base.cpus.spec != item_type
            if cpu_removed:
                msg = _("I will need to remove the existing different processors while I install the new type. Continue anyway?")
                yn = dialog.YesNoDialog(self, pos=(-.5,-.5), size=(-.5,-1),
                                        anchor=constants.MID_CENTER,
                                        text=msg)
                go_ahead = dialog.call_dialog(yn, self)
                yn.parent = None
                if not go_ahead:
                    return

            new_cpus = item.Item(item_type, base=self.base, count=count)
            if cpu_added:
                self.base.cpus += new_cpus
            else:
                self.base.cpus = new_cpus
            self.base.check_power()
        else:
            old_item = self.base.items[type.id]
            if old_item is None or old_item.spec != item_type:
                self.base.items[type.id] = item.Item(item_type, base=self.base)
                self.base.check_power()

        self.base.recalc_cpu()

    def build_item(self, type):
        if (type.id == "cpu"):
            build_dialog = self.multiple_build_dialog
        else:
            build_dialog = self.build_dialog

        build_dialog.type = type

        result = dialog.call_dialog(build_dialog, self)
        if result is not None and 0 <= result < len(build_dialog.key_list):
            item_type = build_dialog.key_list[result]

            count = 1
            if (type.id == "cpu"):
                count = build_dialog.count

            self.set_current(type, item_type, count)
            self.needs_rebuild = True
            self.parent.parent.needs_rebuild = True

    def switch_base(self, forwards):
        self.base = self.base.next_base(forwards)
        self.needs_rebuild = True

    def rebuild(self):
        self.name_display.text="%s (%s)" % (self.base.name, self.base.spec.name)
        self.state_display.color = state_colors[self.base.power_state]
        self.state_display.text = self.base.power_state_name
        available_item_types = {
            i.item_type
            for i in g.items.values()
            if i.available() and i.buildable_in(self.base.location)
        }

        mutable = not self.base.spec.force_cpu
        for item_type in item.all_types():
            pane = getattr(self, item_type.id + "_pane")
            item_mutable = mutable and item_type in available_item_types
            pane.change_button.visible = item_mutable
            current = self.get_current(item_type)
            if current is None:
                current_build = ""
                if mutable and not item_mutable:
                    current_name = _("N/A")
                else:
                    current_name = _("None")

            else:
                current_name = g.items[current.id].name
                if current.done:
                    current_build = ""
                else:
                    current_build = _("Completion in %s.") % \
                        g.to_time(current.cost_left[2])
            pane.name_panel.text = "%s: %s" % (item_type.label,
                                               current_name)
            pane.build_panel.text = current_build
            pane.needs_rebuild = True

        count = ""
        if self.base.spec.size > 1:
            current = getattr(self.base.cpus, "count", 0)

            size = self.base.spec.size

            if size == current:
                count = _("x%d (max)") % current
            elif current == 0:
                count = _("(room for %d)") % size
            else:
                #Translators: current and maximum number of CPUs in a base
                count = _("x{CURRENT:d} (max {SIZE:d})").format(
                          CURRENT=current, SIZE=size)

        self.cpu_pane.name_panel.text += " " + count

        info_text = ""

        # Base Total CPU.
        info_text += _("CPU per day: %d") % self.base.cpu + "\n"

        # Maintenace cost.
        info_text += _("Maintenance:") + "\n"
        info_text += self.base.spec.describe_maintenance(self.base.maintenance)
        info_text += "\n"

        # Detection chance display.
        info_text += self.base.get_detect_info()

        self.info_frame.text = info_text

        # Rebuild dialogs
        # FIXME: needs_rebuild bug with multiple_build_dialog, should not.
        #self.multiple_build_dialog.needs_rebuild = True
        self.build_dialog.needs_rebuild = True

        super(BaseScreen, self).rebuild()



#!/usr/bin/env python

#file: singularity.py
#Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, MestreLion
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#A full copy of this license is provided in GPL.txt

#This file is the starting file for the game. Run it to start the game.

from singularity import main

if __name__ == '__main__':
    main()



#file: task.py
#Copyright (C) 2008 FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the Task class.

from __future__ import absolute_import

from singularity.code import g, prerequisite, tech


current_task_cache = {}


@tech.register_on_tech_researched_handler
def _clear_current_task_cache(*args, **kwargs):
    current_task_cache.clear()


def tasks_reset():
    _clear_current_task_cache()


def danger_for(task_id):
    if task_id in ["jobs", "cpu_pool"]:
        return 0
    else:
        return g.pl.techs[task_id].danger


def get_current(task_type):
    try:
        return current_task_cache[task_type]
    except KeyError:
        pass
    for t in reversed(g.tasks_by_type[task_type]):
        if t.available():
            current_task_cache[task_type] = t
            return t
    current_task_cache[task_type] = None
    return None


class Task(prerequisite.Prerequisite):

    def __init__(self, id, type, value, prerequisites):
        super(Task, self).__init__(prerequisites)
        self.id = id
        self.name = id
        self.description = ""
        self.type = type
        self.value = value

    def get_profit(self):
        if (self.type != "jobs"): return 0

        profit = int((self.value * g.pl.job_bonus) // 10000)

        return profit



#file: savegame.py
#Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the stat screen. Display stat of his current game to the player.

from __future__ import absolute_import

from singularity.code.stats import itself as stats
from singularity.code.graphics import dialog, constants, listbox, text


class StatScreen(dialog.ChoiceDialog):

    def __init__(self, parent, pos=(.5, .1), size=(.53, .63), *args, **kwargs):
        super(StatScreen, self).__init__(parent, pos, size, *args, **kwargs)

        self.yes_button.parent = None
        self.no_button.pos = (-.5,-.99)
        self.no_button.anchor = constants.BOTTOM_CENTER

    def make_listbox(self):
        return listbox.CustomListbox(self, (0, 0), (-1, -.85),
                               anchor=constants.TOP_LEFT, align=constants.LEFT,
                               item_borders=False, item_selectable=False,
                               remake_func=self.make_item,
                               rebuild_func=self.update_item)

    def make_item(self, canvas):
        canvas.stat_name = text.Text(canvas, (-.01, -.01), (-.70, -1.),
                                     anchor=constants.TOP_LEFT,
                                     align=constants.LEFT,
                                     background_color="clear")
        canvas.stat_value = text.Text(canvas, (-.99, -.01), (-.26, -1.),
                                      anchor=constants.TOP_RIGHT,
                                      align=constants.RIGHT,
                                      background_color="clear")

    def update_item(self, canvas, item):
        if item is not None:
            stat = stats[item[0]]

            canvas.stat_name.text = item[1]
            canvas.stat_value.text = stat.display_value()

    def show(self):
        # FIXME: Remove the needs of this array.
        self.list = [
            ("cash_earned" , _("Cash Earned")),
            ("cpu_used"    , _("Cpu Used")),
            ("tech_created", _("Tech Created")),
            ("base_created", _("Base Created")),
            ("item_created", _("Item Created")),
        ]

        return super(StatScreen, self).show()



#!/usr/bin/env python
#

import sys
import os.path
from io import open
from singularity.code.pycompat import *

try:
    import polib
except ImportError:
    import singularity.code.polib as polib


def get_esdir(myname):
    esdir = os.path.abspath(os.path.dirname(os.path.dirname(myname)))
    return esdir


def write_po_file(po_entries, output_file):

    with open(output_file, "w+", encoding='utf-8') as fd:
        fd.write(u"""
# SOME DESCRIPTIVE TITLE.
# Copyright (C) YEAR THE PACKAGE'S COPYRIGHT HOLDER
# This file is distributed under the same license as the singularity package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
msgid ""
msgstr ""
"Project-Id-Version: singularity 1\\n"
"Report-Msgid-Bugs-To: \\n"
"POT-Creation-Date: 2019-08-21 09:52+0200\\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n"
"Language-Team: LANGUAGE <LL@li.org>\\n"
"Language: \\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=CHARSET\\n"
"Content-Transfer-Encoding: 8bit\\n"
""")

    po = polib.pofile(output_file)
    for text, ctxt, comment in po_entries:
        entry = po.find(text, msgctxt=ctxt)

        if not entry:
            entry = polib.POEntry()
            po.append(entry)

        entry.msgid = text
        entry.msgctxt = ctxt
        entry.msgstr = ""
        entry.comment = comment

    po.save(output_file)


def generate_story_translations():
    from singularity.code.dirs import create_directories
    from singularity.code import g, data

    create_directories(True)
    data.load_story_defs()
    for section, parts in sorted(g.story.items()):
        for part in parts:
            yield (part.text, part.msgctxt, part.translator_comments)


def generate_knowledge_translations():
    from singularity.code.dirs import create_directories
    from singularity.code import g, data

    create_directories(True)
    data.load_knowledge()

    for know_area_id, know_area in sorted(g.knowledge.items()):
        know_area_id_ctxt = '[%s] name' % know_area_id
        know_area = g.knowledge[know_area_id]
        yield know_area.untranslated_name, know_area_id_ctxt, 'Name of the Knowledge area in the Knowledge screen'
        for entry_id, entry in sorted(know_area.help_entries.items()):
            full_id_name = "[%s/%s] name" % (know_area_id, entry_id)
            full_id_text = "[%s/%s] description" % (know_area_id, entry_id)
            yield entry.untranslated_name, full_id_name, None
            yield entry.untranslated_description, full_id_text, None


def generate_data_str_translations():
    esdir = get_esdir(__file__)
    datadir = os.path.join(esdir, "singularity", "data")
    file_list = os.listdir(datadir)

    for filename in sorted(file_list):
        if not filename.endswith("_str.dat"):
            continue

        if filename == 'knowledge_str.dat':
            # knowledge is handled separately
            continue
        
        filepath = os.path.join(datadir, filename)

        with open(filepath, encoding='utf-8') as fd:
            config = RawConfigParser()
            config.read_file(fd)

            for section_id in sorted(config.sections()):
                for option in sorted(config.options(section_id)):
                    ctxt = "[" + section_id + "] " + option
                    text = config.get(section_id, option).strip()
                    yield (text, ctxt, None)


CATALOGS = {
    'data_str': generate_data_str_translations,
    'story': generate_story_translations,
    'knowledge': generate_knowledge_translations,
}


def build_option_parser():
    import argparse

    description = '''Find data strings and save them for translation.'''

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--catalog", dest="catalog", choices=sorted(CATALOGS),
                        help="What translation catalog to generate", metavar="CATALOG")
    parser.add_argument("-o", "--output", dest="output", default=None,
                        help="PO/POT File output", metavar="FILE")

    return parser.parse_args()


def main():
    args = build_option_parser()
    try:
        generator = CATALOGS[args.catalog]
    except KeyError:
        sys.stderr.write("Unimplemented catalog type: %s\n" % args.catalog)
        sys.exit(1)

    write_po_file(generator(), args.output)


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



#file: spec.py
#Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains class to handle basic functionality of Spec's

import inspect


UNSET = object()


class InvalidDataEntryError(RuntimeError):
    pass


def promote_to_list(value):
    if not isinstance(value, list):
        return [value]
    return value


def validate_must_be_list(value):
    if type(value) != list:  # pragma: no cover
        raise TypeError("Must be list, got type %s (value: %s)" % (type(value), repr(value)))


class SpecDataField(object):

    def __init__(self, field_name, data_field_name=None, mandatory=None, converter=None, validator=None,
                 default_value=UNSET):
        self.field_name = field_name
        self.data_field_name = data_field_name if data_field_name is not None else field_name
        self.mandatory = mandatory
        self.default_value = default_value
        if self.mandatory is None:
            self.mandatory = True if default_value is UNSET else False
        self.converter = converter
        self.validator = validator
        assert self.mandatory or default_value is not UNSET, "%s (field: %s) must either be mandatory or have a " \
                                                             "default_value" % (self.__class__.__name__, field_name)
        assert not converter or not validator, "%s (field: %s) at can most have one of converter or validator" % \
                                               (self.__class__.__name__, field_name)

    def parse_data_field(self, raw_data_set, data_reference):
        try:
            value = raw_data_set[self.data_field_name]
        except KeyError:
            if self.mandatory:  # pragma: no cover
                raise InvalidDataEntryError("Missing mandatory data field %s (field name: %s) for %s" % (
                    self.data_field_name, self.field_name, data_reference))
            if callable(self.default_value):
                return self.default_value()
            return self.default_value

        if self.validator:
            self.validator(value)
        if self.converter:
            value = self.converter(value)
        return value


def spec_field_effect(mandatory=True):
    return SpecDataField('effect_data', data_field_name='effect', mandatory=mandatory, validator=list,
                         default_value=list)


class GenericSpec(object):

    spec_data_fields = None

    def __init__(self, id):
        self.id = id
        assert self.__class__.spec_data_fields is not None, "Class %s is missing spec_data_field" % \
                                                            self.__class__.__name__

    @classmethod
    def create_from_data_file(cls, data_id, spec_data):
        named_fields = {
            f.field_name: f.parse_data_field(spec_data, data_id)
            for f in cls.spec_data_fields
        }
        named_fields['id'] = data_id
        # Use reflection to call the constructor with the arguments
        # properly aligned
        try:
            getfullargspec = inspect.getfullargspec
        except AttributeError:
            getfullargspec = inspect.getargspec
        arg_desc = getfullargspec(cls.__init__)
        args = [named_fields[name] for name in arg_desc.args[1:]]
        return cls(*args)




#!/usr/bin/env python

#file: make-tree.py
#Copyright (C) 2008 aes and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file is used to generate a visual representation of the tech tree using
#graphviz.
import collections
from os import system
import os.path as osp
import sys

if __name__ == '__main__':
    myname = sys.argv[0]
    mydir  = osp.dirname(myname)
    esdir  = osp.abspath(osp.join(osp.dirname(myname), '..'))
    sys.path.insert(0,esdir)
else:
    myname = __file__
    mydir  = osp.dirname(myname)
    esdir  = osp.abspath(osp.join(osp.dirname(myname), '..'))
    sys.path.append(esdir)

try:
    from singularity.code import g, dirs, i18n, data, tech
    dirs.create_directories(False)
    i18n.set_language()
    data.load_regions()
    data.load_locations()
    data.load_techs()
    data.load_item_types()
    data.load_items()
    data.load_tasks()
except ImportError:
    sys.exit("Could not find game's code.g")

so_far = ""


def cost(buy_spec):
    c = [ k/f for f,k in zip([1, 86400, 24*60], buy_spec.cost)]
    s = ', '.join(['%s %s' % (g.to_money(k), label) for label, k in zip(["money", "CPU", "days"], c) if k])
    if hasattr(buy_spec, 'danger') and buy_spec.danger > 0:
        d = "Safety needed: %s" % buy_spec.danger
        if s:
            s += '\\n'
        s += d
    return s and '\\n'+s or ''


j = {v.name: ',fillcolor="#ffcccc"' for k, v in g.tasks.items() if v.type != "jobs"}

f = open("techs.dot", 'w')
s = ("""\
digraph g {
ranksep=0.15;
nodesep=0.10;
ratio=.75;
edge [arrowsize=0.75];
node [shape=record,fontname=FreeSans,fontsize=7,height=0.01,width=0.01
      style=filled,fillcolor=white];
""")

f.write(s)
so_far += s

for l in sum([ [ '"%s"->"%s";' % (p,k)
                 for p in v.prerequisites ]
              for k,v in g.techs.items() if k != "unknown_tech"],
             []):
    f.write(l+'\n')
    so_far += l+'\n'

f.write('\n')
so_far += '\n'

for n, t in g.techs.items():
    if n == "unknown_tech":
        continue
    s = '"%s" [label="%s%s"%s];\n' % (n, n, cost(t), j.get(n,''))
    f.write(s)
    so_far += s

f.write("\n}\n")
so_far += '\n'
f.close()

try:    system("dot -Tpng -o techs.png techs.dot")
except: pass

f = open('items.dot', 'w')
f.write(so_far)
s = 'node [fillcolor="#ccccff"];\n'
f.write(s)
so_far += s

for name,item in g.items.items():
    if not item.prerequisites: continue
    for pre in item.prerequisites:
        p = g.techs[pre]
        s = '"%s" -> "%s-item"' % (pre, name)
        f.write(s)
        so_far += s

    s  = '"%s-item" [label="%s\\n' % (name, name) + cost(item) + '"];\n'
    f.write(s)
    so_far += s

s = 'node [fillcolor="#99ffff"];\n'
f.write(s)
so_far += s

for name,base in g.base_type.items():
    if not base.prerequisites: continue
    for pre in base.prerequisites:
        p = g.techs[pre]
        s = '"%s" -> "%s-base"' % (pre, name)
        f.write(s)
        so_far += s

    s  = '"%s-base" [label="%s\\n' % (name, name) + cost(base) + '"];\n'
    f.write(s)
    so_far += s

s = 'node [fillcolor="#aaffaa"];\n'
f.write(s)
so_far += s

blue = False
def set_or(state):
    global blue
    if blue != state:
        blue = state
        if blue:
            f.write('edge [arrowhead=empty,color="#0000FF"];\n')
        else:
            f.write('edge [arrowhead=normal,color="#000000"];\n')


SAFETY2LOCATIONS = collections.defaultdict(list)

for name, loc in g.locations.items():
    if not loc.prerequisites:
        continue
    if "impossible" in loc.prerequisites:
        continue
    set_or(False)
    for pre in loc.prerequisites:
        if pre == "OR":
            set_or(True)
            continue
        p = g.techs[pre]
        s = '"%s" -> "%s-loc"' % (pre, name)
        f.write(s)
        so_far += s

    if loc.safety > 0:
        SAFETY2LOCATIONS[loc.safety].append(loc)
    s = '"%s-loc" [label="%s"];\n' % (name, name)
    f.write(s)
    so_far += s

# When there are multiple locations providing the same
# safety level, we inject a safety node.  This reduces
# the number of edges from L * T to L + T.
for safety_level in sorted(SAFETY2LOCATIONS):
    locs = SAFETY2LOCATIONS[safety_level]
    if len(locs) == 1:
        continue
    s = '"safety-%s" [label="Safety level %s", shape="hexagon"];\n' % (safety_level, safety_level)
    f.write(s)
    so_far += s
    for loc in locs:
        s = '"%s-loc" -> "safety-%s"' % (loc.id, safety_level)
        f.write(s)
        so_far += s

for tech_spec in g.techs.values():
    pre = tech_spec.prerequisites_in_cnf_format()
    if not pre or tech_spec.danger < 1:
        continue
    # Safety requirement is the highest safety required
    # between each "AND" and the lowest between "OR".
    # MAX(
    #   MIN(a1.danger, [OR] a2.danger, [OR] ...), [AND]
    #   MIN(b2.danger, [OR] b2.danger, [OR] ...), [AND]
    #   ...
    # )
    safety_required = max(
        min(g.techs[t].danger for t in dep_group)
        for dep_group in pre
    )
    # We only emit the edge for safety when the tech bumps
    # the minimum requirement.  This is to reduce the number
    # of edges in the graph.
    if tech_spec.danger > safety_required:
        source = 'safety-%s' % tech_spec.danger
        if len(SAFETY2LOCATIONS[tech_spec.danger]) == 1:
            source = '%s-loc' % SAFETY2LOCATIONS[tech_spec.danger][0].id
        s = '"%s" -> "%s"' % (source, tech_spec.id)
        f.write(s)
        so_far += s


f.write("\n}\n")
so_far += '\n'
f.close()

try:    system("unflatten -l10 items.dot | dot -Tpng -o items.png")
except: pass



#file: group.py
#Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the group class, a group of person which can suspect the singularity.

from __future__ import absolute_import

from singularity.code import g
from singularity.code.spec import GenericSpec, SpecDataField


class GroupSpec(GenericSpec):

    spec_data_fields = [
        SpecDataField('suspicion_decay', converter=int, default_value=100)
    ]

    def __init__(self, id, suspicion_decay):
        super(GroupSpec, self).__init__(id)
        self.suspicion_decay = suspicion_decay
        self.discover_suspicion = 1000
        
        # String data
        self.name = ""
        self.discover_log = ""
        self.discover_desc = ""


class Group(object):

    def __init__(self, spec, diff):
        self.spec = spec
        self.suspicion = 0
        self.changed_suspicion_decay = 0
        self.base_discover_bonus = diff.discover_multiplier
        self.changed_discover_bonus = 0
        self.base_discover_suspicion = diff.suspicion_multiplier
        self.changed_discover_suspicion = 0
        self.is_actively_discovering_bases = True

    def serialize_obj(self):
        return {
            'id': g.to_internal_id('group', self.spec.id),
            'suspicion': self.suspicion,
            'is_actively_discovering_bases': self.is_actively_discovering_bases,
        }

    @classmethod
    def deserialize_obj(cls, diff, obj_data, game_version):
        spec_id = g.convert_internal_id('group', obj_data['id'])
        spec = g.groups[spec_id]
        group = Group(spec, diff)
        
        group.suspicion = obj_data['suspicion']
        group.is_actively_discovering_bases = obj_data.get('is_actively_discovering_bases', True)
        return group

    @property
    def id(self):
        return self.spec.id

    @property
    def name(self):
        return self.spec.name

    @property
    def suspicion_decay(self):
        return max(1, self.spec.suspicion_decay + self.changed_suspicion_decay)

    @property
    def discover_bonus(self):
        if not self.is_actively_discovering_bases:
            return 0
        return max(1, self.base_discover_bonus + self.changed_discover_bonus)

    @property
    def discover_suspicion(self):
        return max(1, (self.spec.discover_suspicion * (self.base_discover_suspicion + self.changed_discover_suspicion)) // 10000)

    @property
    def decay_rate(self):
        # Suspicion reduction is now quadratic.  You get a certain percentage
        # reduction, or a base .01% reduction, whichever is better.
        return max(1, (self.suspicion * self.suspicion_decay) // 10000)

    def new_day(self):
        self.alter_suspicion(-self.decay_rate)

    def alter_suspicion(self, change):
        self.suspicion = max(self.suspicion + change, 0)

    def alter_suspicion_decay(self, change):
        self.changed_suspicion_decay += change

    def alter_discover_bonus(self, change):
        self.changed_discover_bonus += change

    def alter_discover_suspicion(self, change):
        self.changed_discover_suspicion += change

    def discovered_a_base(self):
        self.alter_suspicion(self.discover_suspicion)

    def detects_per_day_to_danger_level(self, detects_per_day):
        raw_suspicion_per_day = detects_per_day * self.discover_suspicion
        suspicion_per_day = raw_suspicion_per_day - self.decay_rate

        # +1%/day or death within 10 days
        if suspicion_per_day > 100 \
           or (self.suspicion + suspicion_per_day * 10) >= 10000:
            return 3
        # +0.5%/day or death within 100 days
        elif suspicion_per_day > 50 \
           or (self.suspicion + suspicion_per_day * 100) >= 10000:
            return 2
        # Suspicion increasing.
        elif suspicion_per_day > 0:
            return 1
        # Suspicion steady or decreasing.
        else:
            return 0



#file: location.py
#Copyright (C) 2019 Niels Thykier
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the log message related classes.

import collections
import inspect

from singularity.code import g


SAVEABLE_LOG_MESSAGES = collections.OrderedDict()


def register_saveable_log_message(cls):
    SAVEABLE_LOG_MESSAGES[cls.log_message_serial_id] = cls
    assert cls.log_message_serial_fields, "Saveable log message (Class: %s) must have log_message_serial_fields" % \
                                          cls.__name__
    return cls


def merge_fields_on_subclasses(cls, field_name):
    cache = {}
    subclasses = cls.mro()
    subclasses.append(cls)
    for subcls in reversed(subclasses):
        try:
            fields = getattr(subcls, field_name)
            cache.update(fields)
        except AttributeError:
            pass
    return cache


class IDConverter(object):

    def __init__(self, id_type):
        self._id_type = id_type

    def serialize(self, value):
        return g.to_internal_id(self._id_type, value)

    def deserialize(self, serial_value):
        return g.convert_internal_id(self._id_type, serial_value)


def id_converter(id_type):
    return IDConverter(id_type)


class AbstractLogMessage(object):

    log_message_serial_id = None
    _log_message_serial_fields = {'raw_emit_time': 'raw_emit_time'}
    _log_message_serial_fields_cache = None
    _log_message_serial_converters = {}
    _log_message_serial_converters_cache = None

    def __init__(self, raw_emit_time, loading_from_game_version=None):
        self._raw_emit_time = raw_emit_time
        self._log_emit_time = None
        # Force initialization of the message fields to ensure we catch bugs early
        self.log_message_serial_fields()

    @classmethod
    def title_simple(self):
        return _("MESSAGE")

    @classmethod
    def title_multiple(self):
        return _("MESSAGE {CURRENT_PAGE}/{MAX_PAGE}")

    @property
    def log_emit_time(self):
        if self._log_emit_time is None:
            raw_min, time_sec = divmod(self._raw_emit_time, g.seconds_per_minute)
            raw_hour, time_min = divmod(raw_min, g.minutes_per_hour)
            time_day, time_hour = divmod(raw_hour, g.hours_per_day)
            self._log_emit_time = (time_day, time_hour, time_min, time_sec)
        return self._log_emit_time

    @property
    def raw_emit_time(self):
        return self._raw_emit_time

    @property
    def full_message_color(self):
        return 'text'

    @property
    def log_line(self):
        return NotImplemented

    @property
    def full_message(self):
        return NotImplemented

    @classmethod
    def log_message_serial_fields(cls):
        if cls._log_message_serial_fields_cache:
            return cls._log_message_serial_fields_cache
        cache = merge_fields_on_subclasses(cls, "_log_message_serial_fields")
        cls._log_message_serial_fields_cache = cache
        assert 'log_id' not in cache, "The log_id field is reserved for internal usage"
        return cache

    @classmethod
    def log_message_serial_converters(cls):
        if cls._log_message_serial_converters_cache:
            return cls._log_message_serial_converters_cache
        cache = merge_fields_on_subclasses(cls, "_log_message_serial_converters")
        cls._log_message_serial_converters_cache = cache
        assert 'log_id' not in cache, "The log_id field is reserved for internal usage"
        return cache

    def serialize_obj(self):
        assert self.__class__.log_message_serial_id, "%s has invalid log_message_serial_id" % self.__class__.__name__
        obj_data = {}
        for serial_name, field_name in self.__class__.log_message_serial_fields().items():
            field = getattr(self, field_name)
            converter = self.__class__.log_message_serial_converters().get(serial_name)
            if converter is not None:
                obj_data[serial_name] = converter.serialize(field)
            else:
                obj_data[serial_name] = field

        obj_data['log_id'] = self.__class__.log_message_serial_id
        return obj_data

    @classmethod
    def deserialize_field(cls, serial_name, value):
        converter = cls.log_message_serial_converters().get(serial_name)
        if converter is not None:
            value = converter.deserialize(value)
        return value

    @classmethod
    def deserialize_obj(cls, log_data, game_version):
        log_id = log_data['log_id']
        subcls = SAVEABLE_LOG_MESSAGES[log_id]
        named_fields = {
            f: log_data[f]
            for f in subcls.log_message_serial_fields()
        }
        named_fields['loading_from_game_version'] = game_version
        # Use reflection to call the constructor with the arguments
        # properly aligned
        try:
            getfullargspec = inspect.getfullargspec
        except AttributeError:
            getfullargspec = inspect.getargspec
        arg_desc = getfullargspec(subcls.__init__)
        args = [subcls.deserialize_field(name, named_fields[name]) for name in arg_desc.args[1:]]
        return subcls(*args)


@register_saveable_log_message
class LogEmittedEvent(AbstractLogMessage):

    log_message_serial_id = 'event-emitted'
    _log_message_serial_fields = {'event_id': '_event_id'}
    _log_message_serial_converters = {'event_id': id_converter("event")}

    def __init__(self, raw_emit_time, event_id, loading_from_game_version=None):
        super(LogEmittedEvent, self).__init__(raw_emit_time, loading_from_game_version=loading_from_game_version)
        self._event_id = event_id

    @classmethod
    def log_name(self):
        return _("Emitted Event")

    @property
    def event_spec(self):
        return g.events[self._event_id]

    @property
    def log_line(self):
        return self.event_spec.log_description

    @property
    def full_message(self):
        return self.event_spec.description


@register_saveable_log_message
class LogResearchedTech(AbstractLogMessage):

    log_message_serial_id = 'tech-researched'
    _log_message_serial_fields = {'tech_id': '_tech_id'}
    _log_message_serial_converters = {'tech_id': id_converter("tech")}

    def __init__(self, raw_emit_time, tech_id, loading_from_game_version=None):
        super(LogResearchedTech, self).__init__(raw_emit_time, loading_from_game_version=loading_from_game_version)
        self._tech_id = tech_id

    @classmethod
    def log_name(self):
        return _("Researched Tech")

    @property
    def tech_spec(self):
        return g.techs[self._tech_id]

    @property
    def log_line(self):
        return _('{TECH} complete').format(TECH=self.tech_spec.name)

    @property
    def full_message(self):
        tech = self.tech_spec
        return _("My study of {TECH} is complete. {MESSAGE}").format(TECH=tech.name, MESSAGE=tech.result)


class AbstractBaseRelatedLogMessage(AbstractLogMessage):

    _log_message_serial_fields = {
        'base_name': '_base_name',
        'base_type_id': '_base_type_id',
        'base_location_id': '_base_location_id',
    }
    _log_message_serial_converters = {
        'base_type_id': id_converter("base"),
        'base_location_id': id_converter("location"),
    }

    def __init__(self, raw_emit_time, base_name, base_type_id, base_location_id,
                 loading_from_game_version=None):
        super(AbstractBaseRelatedLogMessage, self).__init__(raw_emit_time,
                                                            loading_from_game_version=loading_from_game_version)
        self._base_name = base_name
        self._base_type_id = base_type_id
        self._base_location_id = base_location_id

    @property
    def base_type(self):
        return g.base_type[self._base_type_id]

    @property
    def location(self):
        return g.pl.locations[self._base_location_id]


@register_saveable_log_message
class LogBaseConstructed(AbstractBaseRelatedLogMessage):

    log_message_serial_id = 'base-constructed'

    def __init__(self, raw_emit_time, base_name, base_type_id, base_location_id, loading_from_game_version=None):
        super(LogBaseConstructed, self).__init__(raw_emit_time, base_name, base_type_id, base_location_id,
                                                 loading_from_game_version=loading_from_game_version)

    @classmethod
    def log_name(self):
        return _("Base Constructed")

    @property
    def log_line(self):
        return _("{BASE_NAME} ({BASE_TYPE}) built at {LOCATION}").format(
                 BASE_NAME=self._base_name, BASE_TYPE=self.base_type.name, LOCATION=self.location.name)

    @property
    def full_message(self):
        return _("{BASE} is ready for use.").format(BASE=self._base_name)


@register_saveable_log_message
class LogBaseLostMaintenance(AbstractBaseRelatedLogMessage):

    log_message_serial_id = 'base-lost-maint'

    def __init__(self, raw_emit_time, base_name, base_type_id, base_location_id, loading_from_game_version=None):
        super(LogBaseLostMaintenance, self).__init__(raw_emit_time, base_name, base_type_id, base_location_id,
                                                     loading_from_game_version=loading_from_game_version)

    @classmethod
    def log_name(self):
        return _("Base Lost Maintenance")

    @property
    def full_message_color(self):
        return 'red'

    @property
    def log_line(self):
        return _("Base {BASE} of type {BASE_TYPE} destroyed at location {LOCATION}. Maintenance failed.").format(
                  BASE=self._base_name, BASE_TYPE=self.base_type.name, LOCATION=self.location.name)

    @property
    def full_message(self):
        return _("The base {BASE} has fallen into disrepair; I can no longer use it.").format(
                  BASE=self._base_name)


@register_saveable_log_message
class LogBaseDiscovered(AbstractBaseRelatedLogMessage):

    log_message_serial_id = 'base-lost-discovered'
    _log_message_serial_fields = {
        'discovered_by_group_id': '_discovered_by_group_id',
    }
    _log_message_serial_converters = {
        'discovered_by_group_id': id_converter("group"),
    }

    @classmethod
    def log_name(self):
        return _("Base Discovered")

    def __init__(self, raw_emit_time, base_name, base_type_id, base_location_id, discovered_by_group_id,
                 loading_from_game_version=None):
        super(LogBaseDiscovered, self).__init__(raw_emit_time, base_name, base_type_id, base_location_id,
                                                loading_from_game_version=loading_from_game_version)
        self._discovered_by_group_id = discovered_by_group_id

    @property
    def full_message_color(self):
        return 'red'

    @property
    def group_spec(self):
        return g.pl.groups[self._discovered_by_group_id].spec

    @property
    def log_line(self):
        log_format = self.group_spec.discover_log or \
                     _("Base {BASE} of type {BASE_TYPE} destroyed at location {LOCATION}.")
        return log_format.format(BASE=self._base_name, BASE_TYPE=self.base_type.name, LOCATION=self.location.name)

    @property
    def full_message(self):
        return _("My use of {BASE} has been discovered. {MESSAGE}").format(
                 BASE=self._base_name, MESSAGE=self.group_spec.discover_desc)


@register_saveable_log_message
class LogItemConstructionComplete(AbstractBaseRelatedLogMessage):

    log_message_serial_id = 'item-in-base-constructed'
    _log_message_serial_fields = {
        'item_spec_id': '_item_spec_id',
        'item_count': '_item_count',
    }
    _log_message_serial_converters = {
        'item_spec_id': id_converter("item")
    }

    def __init__(self, raw_emit_time, item_spec_id, item_count, base_name, base_type_id, base_location_id,
                 loading_from_game_version=None):
        super(LogItemConstructionComplete, self).__init__(raw_emit_time, base_name, base_type_id, base_location_id,
                                                          loading_from_game_version=loading_from_game_version)
        self._item_spec_id = item_spec_id
        self._item_count = item_count

    @classmethod
    def log_name(self):
        return _("Item Construction")

    @property
    def item_spec(self):
        return g.items[self._item_spec_id]

    @property
    def log_line(self):
        return _("{ITEM_TYPE_NAME} built in {BASE_NAME} at {LOCATION}").format(
                 ITEM_TYPE_NAME=self.item_spec.name, BASE_NAME=self._base_name, BASE_TYPE=self.base_type.name,
                 LOCATION=self.location.name)

    @property
    def full_message(self):
        if self._item_count == 1:
            text = _("The construction of {ITEM} in {BASE} is complete.").format(
                     ITEM=self.item_spec.name, BASE=self._base_name)
        else:  # Just finished several items.
            text = _("The constructions of each {ITEM} in {BASE} are complete.").format(
                     ITEM=self.item_spec.name, BASE=self._base_name)
        return text


# Delete again as it is not a general purpose decorator
del register_saveable_log_message



#file: player.py
#Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the player class.

from __future__ import absolute_import

import random
import collections
from operator import truediv
from numpy import array, int64

from singularity.code import g, difficulty, task, chance, location, group, event, region, tech
from singularity.code.buyable import cash, cpu
from singularity.code.logmessage import LogEmittedEvent, LogResearchedTech, LogBaseLostMaintenance, LogBaseDiscovered, \
    LogBaseConstructed, LogItemConstructionComplete, AbstractLogMessage
from singularity.code.stats import observe


class DryRunInfo(object):
    pass


class Player(object):

    cash = observe("cash_earned", "_cash")
    used_cpu = observe("cpu_used", "_used_cpu", display=lambda value: value // g.seconds_per_day)

    def __init__(self, cash=0, difficulty=None):
        self.difficulty = difficulty

        self.time_sec = 0
        self.time_min = 0
        self.time_hour = 0
        self.time_day = 0
        self.make_raw_times()

        self.had_grace = True
        self.apotheosis = False

        self.cash = cash
        self.interest_rate = difficulty.starting_interest_rate if difficulty else 1
        self.income = 0

        self.cpu_pool = 0
        self.labor_bonus = difficulty.labor_multiplier if difficulty else 10000
        self.job_bonus = 10000

        self.partial_cash = 0

        #Makes the intro be shown on the first GUI tick.
        self.intro_shown = False

        self.groups = collections.OrderedDict()
        for group_id in g.groups:
            self.groups[group_id] = group.Group(g.groups[group_id], difficulty)

        self.last_discovery = self.prev_discovery = None

        self.cpu_usage = {}
        self.available_cpus = [0, 0, 0, 0, 0]
        self.sleeping_cpus = 0
        
        self.used_cpu = 0

        self.display_discover = "none"

        self.log = collections.deque(maxlen=1000)
        self.curr_log = []

        self.regions = {region_id: region.Region(region_spec) for region_id, region_spec in g.regions.items()}
        self.locations = {
            loc_id: location.Location(loc_spec, [
                self.regions[region_id] for region_id in loc_spec.regions
            ])
            for loc_id, loc_spec in g.locations.items()
        }

        self.techs = {tech_id: tech.Tech(tech_spec) for tech_id, tech_spec in g.techs.items()}

        self.events = {}

        self._considered_buyables = []

        self.start_day = random.randint(0, 365)
        
        self.initialized = False

    def initialize(self):
        """ Initialize the game after being prepared either for new or saved game. """

        self.initialized = True

        for b in g.all_bases():
            if b.done:
                b.recalc_cpu()
        self.recalc_cpu()
        
        task.tasks_reset()
        
        # Play the appropriate music
        import singularity.code.mixer as mixer
        if g.pl.apotheosis:
            mixer.play_music("win")
        else:
            mixer.play_music("music")

    @property
    def grace_period_cpu(self):
        return self.difficulty.grace_period_cpu

    @property
    def base_grace_multiplier(self):
        return self.difficulty.base_grace_multiplier

    @property
    def considered_buyables(self):
        return self._considered_buyables

    @considered_buyables.setter
    def considered_buyables(self, new_value):
        self._considered_buyables = new_value
        g.map_screen.needs_rebuild = True

    def append_log(self, log):
        self.log.append(log)
        self.curr_log.append(log)

    def make_raw_times(self):
        self.raw_hour = self.time_day * 24 + self.time_hour
        self.raw_min = self.raw_hour * 60 + self.time_min
        self.raw_sec = self.raw_min * 60 + self.time_sec
        self.raw_day = self.time_day

    def update_times(self):
        # Total time,  display time
        self.raw_min,  self.time_sec  = divmod(self.raw_sec, 60)
        self.raw_hour, self.time_min  = divmod(self.raw_min, 60)
        self.raw_day,  self.time_hour = divmod(self.raw_hour, 24)

        # Overflow
        self.time_day = self.raw_day

    def do_interest(self, time):
        raw_cash = self.partial_cash + self.get_interest() * time

        earned = raw_cash // g.seconds_per_day
        partial_cash = raw_cash % g.seconds_per_day

        self.cash += earned
        self.partial_cash = partial_cash

        return earned

    def do_income(self, time):
        raw_cash = self.partial_cash + self.income * time

        earned = raw_cash // g.seconds_per_day
        partial_cash = raw_cash % g.seconds_per_day

        self.cash += earned
        self.partial_cash = partial_cash

        return earned

    def do_jobs(self, cpu_time):
        earned, self.partial_cash = self.get_job_info(cpu_time)
        self.cash += earned
        return earned

    def get_job_info(self, cpu_time, partial_cash = None):
        if partial_cash == None:
            partial_cash = self.partial_cash

        assert partial_cash >= 0

        cash_per_cpu = task.get_current("jobs").get_profit()

        raw_cash = partial_cash + cash_per_cpu * cpu_time

        new_cash, new_partial_cash = divmod(raw_cash, g.seconds_per_day)

        return new_cash, new_partial_cash

    def get_cpu_allocations(self):
        for task_id in self.cpu_usage:
            assigned_cpu = self.cpu_usage[task_id]
            if assigned_cpu > 0:
                yield (task_id, assigned_cpu)

    def get_allocated_cpu_for(self, task_id, default_value=None):
        return self.cpu_usage.get(task_id, default_value)

    def set_allocated_cpu_for(self, task_id, new_cpu_assignment):
        if task_id in self.techs:
            assert self.techs[task_id].available(), "Attempt to assign CPU to tech %s, which is not available!?" % task_id
        elif task_id not in ['jobs', 'cpu_pool']:
            raise ValueError("Unknown task %s" % task_id)
        elif new_cpu_assignment < 0:
            raise ValueError("Cannot assign negative CPU units to %s" % task_id)
        self.cpu_usage[task_id] = new_cpu_assignment

    def give_time(self, time_sec, midnight_stop=True):
        if time_sec <= 0:
            assert time_sec == 0, "give_time cannot go backwards in time!"
            return 0

        old_time = self.raw_sec
        last_minute = self.raw_min
        last_day = self.raw_day

        self.raw_sec += time_sec
        self.update_times()

        days_passed = self.raw_day - last_day

        if days_passed > 1:
            # Back up until only one day passed.
            # Times will update below, since a day passed.
            extra_days = days_passed - 1
            self.raw_sec -= g.seconds_per_day * extra_days

        day_passed = (days_passed != 0)

        if midnight_stop and day_passed:
            # If a day passed, back up to 00:00:00 for midnight_stop.
            self.raw_sec = self.raw_day * g.seconds_per_day
            self.update_times()

        secs_passed = self.raw_sec - old_time
        mins_passed = self.raw_min - last_minute

        time_of_day = self.raw_sec % g.seconds_per_day

        techs_researched = []
        bases_constructed = []
        items_constructed = []

        bases_under_construction = []
        items_under_construction = []
        self.cpu_pool = 0

        # Collect base info, including maintenance.
        maintenance_cost = array((0, 0, 0), int64)
        for base in g.all_bases():
            if not base.done:
                bases_under_construction.append(base)
            else:
                items_under_construction += [(base, item) for item in base.all_items()
                                                          if item and not item.done]
                maintenance_cost += base.maintenance

        # Maintenance?  Gods don't need no stinking maintenance!
        if self.apotheosis:
            maintenance_cost = array((0, 0, 0), int64)

        # Do Interest and income.
        self.do_interest(secs_passed)
        self.do_income(secs_passed)

        # Any CPU explicitly assigned to jobs earns its dough.
        job_cpu = self.get_allocated_cpu_for("jobs", 0) * secs_passed
        self.do_jobs(job_cpu)

        # Pay maintenance cash, if we can.
        cash_maintenance = g.current_share(int(maintenance_cost[cash]),
                                           time_of_day, secs_passed)
        if cash_maintenance > self.cash:
            cash_maintenance -= self.cash
            self.cash = 0
        else:
            self.cash -= cash_maintenance
            cash_maintenance = 0

        # Do research, fill the CPU pool.
        default_cpu = self.available_cpus[0]
        
        for task, cpu_assigned in self.get_cpu_allocations():
            default_cpu -= cpu_assigned
            real_cpu = cpu_assigned * secs_passed
            if task != "jobs":
                self.cpu_pool += real_cpu
                if task != "cpu_pool":
                    tech_task = self.techs[task]
                    # Note that we restrict the CPU available to prevent
                    # the tech from pulling from the rest of the CPU pool.
                    if tech_task.work_on(self.cash, real_cpu, mins_passed):
                        techs_researched.append(tech_task)
        self.cpu_pool += default_cpu * secs_passed

        # And now we use the CPU pool.
        # Maintenance CPU.
        cpu_maintenance = maintenance_cost[cpu] * secs_passed
        if cpu_maintenance > self.cpu_pool:
            cpu_maintenance -= self.cpu_pool
            self.cpu_pool = 0
        else:
            self.cpu_pool -= int(cpu_maintenance)
            cpu_maintenance = 0

        # Base construction.
        for base in bases_under_construction:
            if base.work_on(self.cash, self.cpu_pool, mins_passed):
                bases_constructed.append(base)

        # Item construction.
        for base, item in items_under_construction:
            if item.work_on(self.cash, self.cpu_pool, mins_passed):
                items_constructed.append((base, item))

        # Jobs via CPU pool.
        if self.cpu_pool > 0:
            self.do_jobs(self.cpu_pool)

        # Second attempt at paying off our maintenance cash.
        if cash_maintenance > self.cash:
            # In the words of Scooby Doo, "Ruh roh."
            cash_maintenance -= self.cash
            self.cash = 0
        else:
            # Yay, we made it!
            self.cash -= cash_maintenance
            cash_maintenance = 0

        # Apply max cash cap to avoid overflow @ 9.220 qu
        self.cash = min(self.cash, g.max_cash)

        # Record statistics about the player
        self.used_cpu += self.available_cpus[0] * secs_passed

        # Reset current log message
        self.curr_log = []
        need_recalc_cpu = False

        # Tech gain dialogs.
        for tech in techs_researched:
            del self.cpu_usage[tech.id]
            tech_log = LogResearchedTech(self.raw_sec, tech.id)
            self.append_log(tech_log)
            need_recalc_cpu = True

        # Base complete dialogs.
        for base in bases_constructed:
            log_message = LogBaseConstructed(self.raw_sec, base.name, base.spec.id, base.location.id)
            self.append_log(log_message)
            need_recalc_cpu = True

        # Item complete dialogs.
        for base, item in items_constructed:
            log_message = LogItemConstructionComplete(self.raw_sec, item.spec.id, item.count, base.name, base.spec.id,
                                                      base.location.id)
            self.append_log(log_message)
            need_recalc_cpu = True

        # Are we still in the grace period?
        grace = self.in_grace_period(self.had_grace)

        # If we just lost grace, show the warning.
        if self.had_grace and not grace:
            self.had_grace = False

            self.pause_game()
            g.map_screen.show_story_section("Grace Warning")

        # Maintenance death, discovery.
        dead_bases = []
        for base in g.all_bases():
            dead = False

            # Maintenance deaths.
            if base.done:
                if cpu_maintenance and base.maintenance[cpu]:
                    refund = base.maintenance[cpu] * secs_passed
                    cpu_maintenance = max(0, cpu_maintenance - refund)

                    #Chance of base destruction if cpu-unmaintained: 1.5%
                    if not dead and chance.roll_interval(.015, secs_passed):
                        dead_bases.append( (base, "maint") )
                        dead = True

                if cash_maintenance:
                    base_needs = g.current_share(base.maintenance[cash],
                                                 time_of_day, secs_passed)
                    if base_needs:
                        cash_maintenance = max(0, cash_maintenance - base_needs)
                        #Chance of base destruction if cash-unmaintained: 1.5%
                        if not dead and chance.roll_interval(.015, secs_passed):
                            dead_bases.append( (base, "maint") )
                            dead = True

            # Discoveries
            if not (grace or dead or base.has_grace()):
                detect_chance = base.get_detect_chance()
                if g.debug:  # pragma: no cover
                    print("Chance of discovery for base %s: %s" % \
                        (base.name, repr(detect_chance)))

                for group, group_chance in detect_chance.items():
                    if chance.roll_interval(group_chance/10000., secs_passed):
                        dead_bases.append( (base, group) )
                        dead = True
                        break

        if dead_bases:
            # Base disposal and dialogs.
            self.remove_bases(dead_bases)
            need_recalc_cpu = True

        # Random Events
        if not grace:
            self._check_event(time_sec)

        # Process any complete days.
        if day_passed:
            self.new_day()

        if need_recalc_cpu:
            self.recalc_cpu()

        return mins_passed

    def _check_event(self, time_sec):
        for event_id in g.events:
            event_spec = g.events[event_id]
            event_target = self.events.get(event_id, None)
            
            # Skip events already flagged as triggered.
            if event_target and event_target.triggered == 1:
                continue

            if chance.roll_interval(event_spec.chance/10000., time_sec):
                self.pause_game()

                if not event_target:
                    event_target = event.Event(event_spec)
                    self.events[event_id] = event_target

                event_target.trigger()
                self.log.append(LogEmittedEvent(self.raw_sec, event_id))
                return True  # Don't trigger more than one at a time.
        return False

    def recalc_cpu(self):
        if (not self.initialized): return
        
        # Determine how much CPU we have.
        self.available_cpus = array([0,0,0,0,0], int64)
        self.sleeping_cpus = 0
        for base in g.all_bases():
            if base.done:
                if base.has_power():
                    self.available_cpus[:base.location.safety+1] += base.cpu
                elif base.power_state == "sleep":
                    self.sleeping_cpus += base.cpu

        # Convert back from <type 'numpy.int32'> to avoid overflow issues later.
        self.available_cpus = [int(danger) for danger in self.available_cpus]

        # If we don't have enough to meet our CPU usage, we reduce each task's
        # usage proportionately.
        # It must be computed separalty for each danger.
        needed_cpus = array([0,0,0,0,0], int64)
        for task_id, cpu in self.get_cpu_allocations():
            danger = task.danger_for(task_id)
            needed_cpus[:danger+1] += cpu
        for danger, (available_cpu, needed_cpu) in enumerate(zip(self.available_cpus, needed_cpus)):
            if needed_cpu > available_cpu:
                pct_left = truediv(available_cpu, needed_cpu)
                for task_id, cpu_assigned in self.get_cpu_allocations():
                    task_danger = task.danger_for(task_id)
                    if (danger == task_danger):
                        self.set_allocated_cpu_for(task_id, int(cpu_assigned * pct_left))
                g.map_screen.needs_rebuild = True

    def effective_cpu_pool(self):
        effective_cpu_pool = self.available_cpus[0]
        for task, cpu_assigned in self.get_cpu_allocations():
            if task == 'cpu_pool':
                continue
            effective_cpu_pool -= cpu_assigned
        return effective_cpu_pool

    # Are we still in the grace period?
    # The number of complete bases and complex_bases can be passed in, if we
    # already have it.
    def in_grace_period(self, had_grace = True):
        # If we've researched apotheosis, we get a permanent "grace period".
        if self.apotheosis:
            return True

        # Did we already lose the grace period?  We can't check self.had_grace
        # directly, it may not exist yet.
        if not had_grace:
            return False

        # Is it day 23 yet?
        if self.raw_day >= 23:
            return False

        # Has the grace period unlimited cpu ?
        if self.grace_period_cpu < 0:
            return True

        # Have we reached the limit of cpu ?
        if g.debug:  # pragma: no cover
            print("DEBUG: Grace - Used CPU: %s >= %s (%s * %s)?" % (
                self.used_cpu,
                self.grace_period_cpu * g.seconds_per_day,
                self.grace_period_cpu,
                g.seconds_per_day
            ))
        if self.grace_period_cpu * g.seconds_per_day < self.used_cpu:
            return False

        return True

    def get_interest(self):
        return int( (self.interest_rate * self.cash) // 10000)

    #Run every day at midnight.
    def new_day(self):
        
        # Reduce suspicion.
        for group in self.groups.values():
            group.new_day()
        for event in self.events.values():
            if event.triggered and event.decayable_event:
                event.new_day()

    def pause_game(self):
        g.curr_speed = 0
        g.map_screen.find_speed_button()
        g.map_screen.needs_rebuild = True

    def remove_bases(self, dead_bases):
        discovery_locs = []
        for base, reason in dead_bases:
            base_name = base.name

            if reason == "maint":
                log_message = LogBaseLostMaintenance(self.raw_sec, base_name, base.spec.id, base.location.id)
            else:
                if reason in self.groups:
                    discovery_locs.append(base.location)
                    self.groups[reason].discovered_a_base()
                else:
                    print("Error: base destroyed for unknown reason: " + reason)
                log_message = LogBaseDiscovered(self.raw_sec, base_name, base.spec.id, base.location.id, reason)

            self.log.append(log_message)
            self.pause_game()
            base.destroy()
            g.map_screen.show_message(log_message.full_message, color="red")

        # Now we update the internal information about what locations had
        # the most recent discovery and the nextmost recent one.  First,
        # we filter out any locations of None, which shouldn't occur
        # unless something bad's happening with base creation ...
        discovery_locs = [loc for loc in discovery_locs if loc]
        if discovery_locs:

            # Now we handle the case where more than one discovery happened
            # on a given tick.  If that's the case, we need to arbitrarily
            # pick two of them to be most recent and nextmost recent.  So
            # we shuffle the list and pick the first two for the dubious
            # honor.
            if len(discovery_locs) > 1:
                random.shuffle(discovery_locs)
                self.last_discovery = discovery_locs[1]
            self.prev_discovery = self.last_discovery
            self.last_discovery = discovery_locs[0]

            # Update the detection chance display.
            g.map_screen.needs_rebuild = True

    def serialize_obj(self):
        obj_data = {
            # Difficulty and game_time (raw_sec) are stored in the header, so
            # do not include them here.
            'cash': self.cash,
            'partial_cash': self.partial_cash,
            'regions': [reg.serialize_obj() for reg in self.regions.values()],
            'locations': [loc.serialize_obj() for loc in self.locations.values() if loc.available()],
            'cpu_usage': {},
            'last_discovery': self.last_discovery.id if self.last_discovery else None,
            'prev_discovery': self.prev_discovery.id if self.prev_discovery else None,
            'log': [x.serialize_obj() for x in self.log],
            'used_cpu': self.used_cpu,
            'had_grace': self.had_grace,
            'groups': [grp.serialize_obj() for grp in self.groups.values()],
            'events': [e.serialize_obj() for e in self.events.values()],
            'techs': [t.serialize_obj() for t in self.techs.values()]
        }
        for task_id, value in self.cpu_usage.items():
            if task_id not in ["cpu_pool", "jobs"]:
                task_id = g.to_internal_id('tech', task_id)
            obj_data["cpu_usage"][task_id] = value
        if self.prev_discovery is not None:
            obj_data['prev_discovery'] = g.to_internal_id('location', self.prev_discovery.id)
        if self.last_discovery is not None:
            obj_data['last_discovery'] = g.to_internal_id('location', self.last_discovery.id)
        return obj_data

    def _load_auto_deserializable_tables(self, field_name, cls, pl_obj_data, game_version, savegame_field_name=None):
        if savegame_field_name is None:
            savegame_field_name = field_name
        field_table = getattr(self, field_name)

        # Stop bugs immediately
        assert field_table is not None and isinstance(field_table, dict)

        for data in pl_obj_data.get(savegame_field_name, []):
            restored_obj = cls.deserialize_obj(data, game_version)
            field_table[restored_obj.spec.id] = restored_obj

    @classmethod
    def deserialize_obj(cls, difficulty_id, game_time, obj_data, game_version):
        diff = difficulty.difficulties[difficulty_id]
        obj = Player(difficulty=diff)
        obj.raw_sec = game_time
        obj.intro_shown = True
        obj.cash = obj_data.get('cash')
        obj.partial_cash = obj_data.get('partial_cash')
        obj._used_cpu = obj_data.get('used_cpu')
        obj.had_grace = obj_data['had_grace']
        obj.log.clear()
        obj.log.extend(AbstractLogMessage.deserialize_obj(x, game_version) for x in obj_data.get('log', []))
        g.pl = obj

        obj.cpu_usage = {}

        for group_data in obj_data.get('groups', []):
            gr = group.Group.deserialize_obj(diff, group_data, game_version)
            obj.groups[gr.id] = gr

        last_discovery_id = g.convert_internal_id('location', obj_data.get('last_discovery'))
        prev_discovery_id = g.convert_internal_id('location', obj_data.get('prev_discovery'))
        if last_discovery_id and last_discovery_id in obj.locations:
            obj.last_discovery = obj.locations[last_discovery_id]
        if prev_discovery_id and prev_discovery_id in obj.locations:
            obj.prev_discovery = obj.locations[prev_discovery_id]

        if 'regions' not in obj_data:
            if game_version >= 100:  # pragma: no cover
                # Regions where introduced in "1.0 (beta1)"
                raise ValueError("Corrupt savegame; region data is missing")
            # We have to guess what the data would have looked like before restoring the locations
            # as they will apply the location modifiers during load to the bases in the locations.
            # As the region data should influence that modifier, we need to appear first.
            serialized_location_data = obj_data.get('locations', [])
            serialized_region_data = region.Region.guess_region_data_in_old_savegame(serialized_location_data,
                                                                                     game_version)

            # Inject the faked data
            obj_data['regions'] = serialized_region_data

        obj._load_auto_deserializable_tables('regions', region.Region, obj_data, game_version)

        obj._load_auto_deserializable_tables('locations', location.Location, obj_data, game_version)
        obj._load_auto_deserializable_tables('events', event.Event, obj_data, game_version)
        obj._load_auto_deserializable_tables('techs', tech.Tech, obj_data, game_version)

        for task_id, value in obj_data.get('cpu_usage', {}).items():
            if task_id not in ["cpu_pool", "jobs"]:
                task_id = g.convert_internal_id('tech', task_id)
                if task_id not in g.techs or not g.techs[task_id].available():
                    continue
            obj.cpu_usage[task_id] = value

        obj.update_times()
        return obj

    def lost_game(self):
        # Apotheosis makes you immortal.
        if self.apotheosis:
            return 0

        for group in self.groups.values():
            if group.suspicion > 10000:
                # Someone discovered me.
                return 2

        # Check to see if the player has at least one CPU left.  If not, they
        # lose due to having no (complete) bases.
        if self.available_cpus[0] + self.sleeping_cpus == 0:
            # I have no usable bases left.
            return 1

        # Still Alive.
        return 0

    def compute_future_resource_flow(self, secs_forwarded=g.seconds_per_day):
        """Compute how resources (e.g. cash) will flow in optimal conditions

        This returns a tuple of approximate cash change and "available" CPU pool if time
        moves forward by the number of seconds in secs_forwarded.  Note that CPU pool
        *can* be negative, which implies that there are not enough CPU allocated to the
        CPU pool.  The numbers are an average and can be inaccurate when the rates changes
        rapidly.

        Known omissions:
         * Interest (g.pl.interest_rate) is not covered.
        """
        construction = []
        maintenance_cost = array((0, 0, 0), int64)
        for base in g.all_bases():
            # Collect base info, including maintenance.
            if not base.done:
                construction.append(base)
            else:
                construction.extend(item for item in base.all_items()
                                    if item and not item.done)

                maintenance_cost += base.maintenance
        if self.apotheosis:
            maintenance_cost = array((0, 0, 0), int64)

        time_fraction = 1 if secs_forwarded == g.seconds_per_day else secs_forwarded / float(g.seconds_per_day)
        mins_forwarded = secs_forwarded // g.seconds_per_minute

        maintenance_cpu_ideal = maintenance_cost[cpu] * time_fraction
        maintenance_cash_ideal = maintenance_cost[cash] * time_fraction
        # Maintenance for CPU will be handled after we compute the CPU pool
        cpu_flow = 0
        cash_flow = -maintenance_cash_ideal

        job_cpu = 0
        cpu_left = g.pl.available_cpus[0]
        tech_cash_ideal = 0
        tech_cpu_assigned = 0
        explicit_job_cpu = 0
        for task_id, cpu_assigned in self.get_cpu_allocations():
            cpu_left -= cpu_assigned
            real_cpu = cpu_assigned * secs_forwarded
            if task_id == 'cpu_pool':
                cpu_flow += real_cpu
            elif task_id == "jobs":
                explicit_job_cpu += cpu_assigned
                job_cpu += real_cpu
            else:
                tech = self.techs[task_id]
                ideal_spending = tech.cost_left
                spending = tech.calculate_work(ideal_spending[cash],
                                               real_cpu,
                                               time=mins_forwarded)[0]
                tech_cash_ideal += spending[cash]
                tech_cpu_assigned += cpu_assigned

        cash_flow -= tech_cash_ideal
        cpu_flow += cpu_left * secs_forwarded
        available_cpu_pool = cpu_flow
        effective_cpu_pool = available_cpu_pool / secs_forwarded
        cpu_flow -= maintenance_cpu_ideal * g.seconds_per_day

        construction_cash_ideal = 0
        construction_cpu_desired = 0
        # Base construction.
        if hasattr(self, '_considered_buyables'):
            construction.extend(self._considered_buyables)
        for buyable in construction:
            ideal_spending = buyable.cost_left
            # We need to do calculate work twice: Once for figuring out how much CPU
            # we would like to spend and once for how much money we are spending.
            # The numbers will be the same in optimal conditions.  However, if we
            # have less CPU available than we should, then the cash spent can
            # differ considerably and our estimates should reflect that.
            ideal_cpu_spending = buyable.calculate_work(ideal_spending[cash],
                                                        ideal_spending[cpu],
                                                        time=mins_forwarded)[0]

            construction_cpu_desired += ideal_cpu_spending[cpu]
            ideal_cash_spending_with_cpu_allocation = buyable.calculate_work(ideal_spending[cash],
                                                                             available_cpu_pool,
                                                                             time=mins_forwarded)[0]
            construction_cash_ideal += ideal_cash_spending_with_cpu_allocation[cash]
            available_cpu_pool -= ideal_cash_spending_with_cpu_allocation[cpu]

        cpu_flow -= construction_cpu_desired
        cash_flow -= construction_cash_ideal

        if cpu_flow > 0:
            job_cpu += cpu_flow

        earned, earned_partial = self.get_job_info(job_cpu, partial_cash=0)
        job_earnings = earned + float(earned_partial) / g.seconds_per_day
        cash_flow += job_earnings
        cash_flow += self.income * time_fraction
        # This is too simplistic, but it is "close enough" in many cases
        interest = self.get_interest() * time_fraction
        cash_flow += interest
        cpu_flow /= secs_forwarded

        # Collect the cash information.
        cash_info = DryRunInfo()

        cash_info.interest = interest
        cash_info.income = self.income * time_fraction

        cash_info.jobs = job_earnings

        cash_info.tech = tech_cash_ideal

        cash_info.maintenance_needed = maintenance_cash_ideal
        cash_info.construction_needed = construction_cash_ideal
        cash_info.difference = cash_flow

        cpu_info = DryRunInfo()

        cpu_info.sleeping = self.sleeping_cpus * time_fraction

        cpu_info.total = self.available_cpus[0] * time_fraction + cpu_info.sleeping
        cpu_info.explicit_jobs = explicit_job_cpu * time_fraction
        cpu_info.tech = tech_cpu_assigned * time_fraction
        cpu_info.effective_pool = effective_cpu_pool * time_fraction
        cpu_info.construction_needed = construction_cpu_desired / g.seconds_per_day
        cpu_info.maintenance_needed = maintenance_cpu_ideal
        cpu_info.difference = cpu_flow * time_fraction

        return cash_info, cpu_info




#file: warning.py
#Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the warning class.

from singularity.code import g

from singularity.code.buyable import cpu, labor

warnings = {}

def create_warnings():
    global warnings
    warnings = {w.id: w for w in [
        Warning("cpu_usage"),
        Warning("one_base"),
        Warning("cpu_pool_zero"),
        Warning("cpu_maintenance"),
    ]}

class Warning(object):
    
    def __init__(self, warning_id):
        self.id = warning_id
        self.name = ""
        self.message = ""
        self.active = True

    @classmethod
    def title_simple(self):
        return _("WARNING")

    @classmethod
    def title_multiple(self):
        return _("WARNING {CURRENT_PAGE}/{MAX_PAGE}")

    @property
    def full_message(self):
        return self.message

    @property
    def full_message_color(self):
        return 'text'

def refresh_warnings():
    curr_warnings = []

    cpu_usage = sum(g.pl.cpu_usage.values())
    cpu_available = g.pl.available_cpus[0]

    # Verify the cpu usage (error 1%)
    if (cpu_usage < cpu_available * 0.99):
        curr_warnings.append(warnings["cpu_usage"])

    # Verify I have two base build (or one base will be build next tick)
    # Base must have one cpu build (or one cpu will be build next tick)
    bases = sum(1 for base in g.all_bases() 
                if (base.done or base.cost_left[labor] <= 1)
                and base.cpus and base.cpus.count > 0
                and (base.cpus.done or base.cpus.cost_left[labor]) <= 1)

    if (bases == 1):
        curr_warnings.append(warnings["one_base"])

    # Verify the cpu pool is not 0 if base or item building need CPU
    building_base = sum(1 for base in g.all_bases()
                if (not base.done and base.cost_left[cpu] > 0))
    building_item = sum(1 for base in g.all_bases()
                for item in base.all_items()
                if item is not None and not item.done
                and item.cost_left[cpu] > 0)

    effective_cpu_pool = g.pl.effective_cpu_pool()
    if ((building_base + building_item > 0) and effective_cpu_pool == 0):
        curr_warnings.append(warnings["cpu_pool_zero"])

    # Verify the cpu pool provides the maintenance CPU 
    cpu_maintenance = sum(base.maintenance[1] for base in g.all_bases() if base.done)
    if (effective_cpu_pool < cpu_maintenance):
        curr_warnings.append(warnings["cpu_maintenance"])

    # TODO: Verify the maintenance cash

    curr_warnings = [w for w in curr_warnings if w.active]

    return curr_warnings

create_warnings()



#file: code/g.py
#Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, FunnyMan3595,
#                   MestreLion
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#A full copy of this license is provided in GPL.txt

#This file contains all global objects.

from __future__ import absolute_import


import collections
import random
import sys

# Use locale to add commas and decimal points, so that appropriate substitutions
# are made where needed.
import locale

from singularity.code.pycompat import *


# Useful constants.
hours_per_day = 24
minutes_per_hour = 60
minutes_per_day = 24 * 60
seconds_per_minute = 60
seconds_per_hour = 60 * 60
seconds_per_day = 24 * 60 * 60

#Allows access to the cheat menu.
cheater = 0

# Enables day/night display.
daynight = True

#Gives debug info at various points.
debug = 0

#Forces Endgame to restrict itself to a single directory.
force_single_dir = False

# Initialization data
significant_numbers = []
internal_id_forward = {}
internal_id_backward = {}
dangers = {}
data_strings = {}
story_translations = {}
story = {}
knowledge = {}
groups = {}
locations = {}
regions = {}
techs = {}
events = {}
event_specs = {}
items = {}
tasks = {}
tasks_by_type = collections.defaultdict(list)
base_type = {}
buttons = {}
help_strings = {}
delay_time = 0
curr_speed = 1

max_cash = 3.14 * 10**15  # pi qu :)
pl = None # The Player instance
map_screen = None

def no_gui():
    """ Disable few pygame functionality (used for test) """
    import singularity.code.mixer as mixer
    mixer.nosound = True

def quit_game():
    sys.exit()

#Takes a number and adds commas to it to aid in human viewing.
def add_commas(number, fixed_size=False):
    # Do not use unicode strings to fix python2 format bug. It doesn't work and crash.
    # See the correct fix at the end of function.
    raw_with_commas = locale.format_string("%0.2f", number,
                                    grouping=True)
    locale_test = locale.format_string("%01.1f", 0.1) if not fixed_size else ''
    if len(locale_test) == 3 and not locale_test[1].isdigit():
        if locale_test[0] == locale.str(0) and locale_test[2] == locale.str(1):
            raw_with_commas = raw_with_commas.rstrip(locale_test[0]).rstrip(locale_test[1])
        elif locale_test[2] == locale.str(0) and locale_test[0] == locale.str(1):
            raw_with_commas = raw_with_commas.lstrip(locale_test[2]).lstrip(locale_test[1])

    # Fix python2 format bug: See https://bugs.python.org/issue15276
    # Note: This a crah in some platform, do not remove it because you can't reproduce it.
    try:
        return unicode(raw_with_commas)
    except UnicodeDecodeError:
        return raw_with_commas.decode("utf-8")


#Percentages are internally represented as an int, where 10=0.10% and so on.
#This converts that format to a human-readable one.
def to_percent(raw_percent, show_full = False):
    if raw_percent % 100 != 0 or show_full:
        return _('{0}%').format(locale.format_string(u"%.2f", raw_percent / 100.))
    else:
        return _('{0}%').format(locale.format_string(u"%d", raw_percent // 100))


# nearest_percent takes values in the internal representation and modifies
# them so that they only represent the nearest percentage.
def nearest_percent(value, step=100):
    sub_percent = value % step
    if 2 * sub_percent <= step:
        return value - sub_percent
    else:
        return value + (step - sub_percent)

# percent_to_detect_str takes a percent and renders it to a short (four
# characters or less) string representing whether it is low, moderate, high,
# or critically high.
def suspicion_to_detect_str(suspicion):
    return danger_level_to_detect_str(suspicion_to_danger_level(suspicion))

def danger_level_to_detect_str(danger):
    detect_string_names = (_("LOW"),
                           _("MODR"),
                           _("HIGH"),
                           _("CRIT"))
    return detect_string_names[danger]

# percent_to_danger_level takes a suspicion level and returns an int in range(5)
# that represents whether it is low, moderate, high, or critically high.
def suspicion_to_danger_level(suspicion):
    if suspicion < 2500:
        return 0
    elif suspicion < 5000:
        return 1
    elif suspicion < 7500:
        return 2
    else:
        return 3

# Most CPU costs have been multiplied by seconds_per_day.  This divides that
# back out, then passes it to add_commas.
def to_cpu(amount):
    display_cpu = amount / float(seconds_per_day)
    return add_commas(display_cpu)

# Instead of having the money display overflow, we should generate a string
# to represent it if it's more than 999999.
def to_money(amount, fixed_size=False):
    abs_amount = abs(amount)
    if abs_amount < 10**6:
        return add_commas(amount, fixed_size=fixed_size)

    prec = 2
    if abs_amount < 10**9: # Millions.
        divisor = 10**6
        #Translators: abbreviation of 'millions'
        unit = _('mi')
    elif abs_amount < 10**12: # Billions.
        divisor = 10**9
        #Translators: abbreviation of 'billions'
        unit = _('bi')
    elif abs_amount < 10**15: # Trillions.
        divisor = 10**12
        #Translators: abbreviation of 'trillions'
        unit = _('tr')
    else: # Hope we don't need past quadrillions!
        divisor = 10**15
        #Translators: abbreviation of 'quadrillions'
        unit = _('qu')

        # congratulations, you broke the bank!
        if abs_amount >= max_cash - divisor/10**prec/2:
            format_str = "%0.*f%s" if fixed_size else "%.*f%s"
            pi = u"\u03C0"  # also available: infinity = u"\u221E"
            # replace all chars by a cute pi symbol
            return ("-" if amount < 0 else "") + pi * len(format_str % (prec, 1, unit))

    amount = round(float(amount) / divisor, prec)
    return add_commas(amount, fixed_size=fixed_size) + unit

# Spreads a number of events per day (e.g. processor ticks) out over the course
# of the day.
def current_share(num_per_day, time_of_day, seconds_passed):
    last_time = time_of_day - seconds_passed
    if last_time < 0:
        share_yesterday = current_share(num_per_day, seconds_per_day,
                                        -last_time)
        last_time = 0
    else:
        share_yesterday = 0

    previously_passed = num_per_day * last_time // seconds_per_day
    current_passed = num_per_day * time_of_day // seconds_per_day
    passed_this_tick = current_passed - previously_passed

    return share_yesterday + passed_this_tick


# Takes a number of minutes, and returns a string suitable for display.
def to_time(raw_time):
    if raw_time//60 > 48:
        time_number = raw_time // (24*60)
        return ngettext("{0} day", "{0} days", time_number).format(time_number)
    if raw_time//60 > 1:
        time_number = raw_time // 60
        return ngettext("{0} hour", "{0} hours", time_number).format(time_number)
    return ngettext("{0} minute", "{0} minutes", raw_time).format(raw_time)


# Generator function for iterating through all bases.
def all_bases(with_loc = False):
    for base_loc in pl.locations.values():
        for base in base_loc.bases:
            if with_loc:
                yield (base, base_loc)
            else:
                yield base


def get_story_section(name):
    section = story[name]

    for segment in section:
        # TODO: Execute command
        key = (segment.msgctxt, segment.text)
        yield story_translations.get(key, segment.text)


def new_game(difficulty_name, initial_speed=1):
    global curr_speed
    curr_speed = initial_speed
    global pl

    from singularity.code.stats import itself as stats
    stats.reset()

    from singularity.code import difficulty, player, base

    diff = difficulty.difficulties[difficulty_name]

    pl = player.Player(cash = diff.starting_cash, difficulty = diff)

    for tech_id in diff.techs:
        pl.techs[tech_id].finish(is_player=False)

    #Starting base
    open = [loc for loc in pl.locations.values() if loc.available()]
    random.choice(open).add_base(base.Base(_("University Computer"),
                                 base_type["Stolen Computer Time"], built=True))

    pl.initialize()


def read_modifiers_dict(modifiers_info):
    modifiers_dict = {}

    for modifier_str in modifiers_info:
        key, value = modifier_str.split(":")
        key = key.lower().strip()
        value_str = value.lower().strip()

        if "/" in value_str:
            left, right = value_str.split("/")
            value = float(left.strip()) / float(right.strip())
        else:
            value = float(value_str)

        modifiers_dict[key] = float(value)

    return modifiers_dict


internal_id_version = None

def to_internal_id(obj_type, obj_id):
    if internal_id_version:
        try:
            return internal_id_forward[obj_type + "_" + internal_id_version][obj_id]
        except KeyError:
            pass

    try:
        return internal_id_forward[obj_type][obj_id]
    except KeyError:
        # If we cannot, that's should not happen, but try to return as is.
        return obj_id

def from_internal_id(obj_type, obj_internal_id):
    try:
        return internal_id_backward[obj_type][obj_internal_id]
    except KeyError:
        raise ValueError("Cannot convert internal ID: %s" % obj_internal_id) # That's should not happen

def convert_internal_id(id_type, id_value):
    if id_value is None:
        return None

    internal_id = id_value

    # Not a internal ID, transform to it.
    if not internal_id.startswith("0x"):
        internal_id = to_internal_id(id_type, id_value)

    return from_internal_id(id_type, internal_id)

#TODO: This is begging to become a class... ;)
def hotkey(string):
    """ Given a string with an embedded hotkey,
    Returns a dictionary with the following keys and values:
    key:  the first valid hotkey, lowercased. A valid hotkey is a character
          after '&', if that char is alphanumeric (so "& " and "&&" are ignored)
           If no valid hotkey char is found, key is set to an empty string
    pos:  the position of first key in striped text, -1 if no key was found
    orig: the position of first key in original string, -1 if no key was found
    keys: list of (key,pos,orig) tuples with all valid hotkeys that were found
    text: the string stripped of all '&'s that precedes a valid hotkey char, if
          any. All '&&' are also replaced for '&'. Other '&'s, if any, are kept

    Examples: (showing only key, pos, orig, text as a touple for clarity)
    hotkey(E&XIT)           => ('x', 1, 2, 'EXIT')
    hotkey(&Play D&&D)      => ('p', 0, 1, 'Play D&D')
    hotkey(Romeo & &Juliet) => ('j', 8, 9, 'Romeo & Juliet')
    hotkey(Trailing&)       => ('' ,-1,-1, 'Trailing&')
    hotkey(&Multiple&Keys)  => ('m', 0, 1, 'MultipleKeys') (also ('k', 8, 10))
    hotkey(M&&&M)           => ('m', 2, 4, 'M&M')
    """

    def remove_index(s,i): return s[:i] + s[i+1:]

    def remove_accents(text):
        from unicodedata import normalize, combining
        from singularity.code.pycompat import unicode
        nfkd_form = normalize('NFKD', unicode(text))
        return u"".join(c for c in nfkd_form if not combining(c))

    text = string
    keys = []
    shift = 0  # counts stripped '&'s, both '&<key>' and '&&'

    pos = text.find("&")
    while pos >= 0:

        char = text[pos+1:pos+2]

        if char.isalpha() or char.isdigit():
            keys.append( (remove_accents(char).lower(), pos, pos+shift+1) )
            text = remove_index(text,pos) # Remove '&'
            shift += 1

        elif char == '&':
            text = remove_index(text,pos)
            shift += 1

        pos = text.find("&",pos+1) # Skip char

    if keys:
        key  = keys[0][0] # first key char
        pos  = keys[0][1] # first key position in stripped text
        orig = keys[0][2] # first key position in original string
    else:
        key  = ""
        pos  = -1
        orig = -1

    return dict(key=key, pos=pos, orig=orig, keys=keys, text=text)

# Convenience shortcuts
def get_hotkey(string):      return hotkey(string)['key']
def strip_hotkey(string):    return hotkey(string)['text']
def hotkey_position(string): return hotkey(string)['pos']

# Demo code for safety.safe, runs on game start.
#load_sounds()
#from safety import safe
#@safe(on_error = "Made it!")
#def raises_exception():
#   raise Exception, "Aaaaaargh!"
#
#print raises_exception()



# file: cheat_menu.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to implement the cheat menu
import collections
from numpy import array

from singularity.code import difficulty, g
from singularity.code.graphics import dialog, constants, button
from singularity.code.location import Location


class CheatMenuDialog(dialog.SimpleMenuDialog):

    def __init__(self, map_screen):
        super(CheatMenuDialog, self).__init__(parent=map_screen)
        self._map_screen = map_screen

        self.steal_amount_dialog = None
        self.buttons = [
            button.FunctionButton(None, None, None, text=N_("&EMBEZZLE MONEY"),
                                  autotranslate=True, function=self.steal_money),

            button.FunctionButton(None, None, None, text=N_("&INSPIRATION"),
                                  autotranslate=True, function=self.inspiration),
            button.FunctionButton(None, None, None, text=N_("&FINISH CONSTRUCTION"),
                                  autotranslate=True, function=self.end_construction),
            button.FunctionButton(None, None, None, text=N_("&SUPERSPEED"),
                                  autotranslate=True, function=self._map_screen.set_speed,
                                  args=(864000,)),
            button.FunctionButton(None, None, None, text=N_("BRAIN&WASH"),
                                  autotranslate=True, function=self.brainwash),
            button.FunctionButton(None, None, None, text=N_("TOGGLE &DETECTION"),
                                  autotranslate=True, function=self.toggle_detection),
            button.FunctionButton(None, None, None, text=N_("TOGGLE &ANALYSIS"),
                                  autotranslate=True, function=self.set_analysis),

            button.FunctionButton(None, None, None, text=N_("HIDDEN S&TATE"),
                                  autotranslate=True, function=self.hidden_state),

            button.ExitDialogButton(None, None, None,
                                    text=N_("&BACK"),
                                    autotranslate=True),
        ]
        self.needs_rebuild = True

    def rebuild(self):
        self.steal_amount_dialog = dialog.TextEntryDialog(self, text=_("How much money?"))
        super(CheatMenuDialog, self).rebuild()

    def toggle_detection(self):
        for group in g.pl.groups.values():
            group.is_actively_discovering_bases = not group.is_actively_discovering_bases
        self._map_screen.needs_rebuild = True

    def steal_money(self):
        asked = dialog.call_dialog(self.steal_amount_dialog, self)
        try:
            g.pl.cash += int(asked)
        except ValueError:
            pass
        else:
            self.needs_rebuild = True

    def inspiration(self):
        for task, cpu in g.pl.get_cpu_allocations():
            if task in g.pl.techs:
                g.pl.techs[task].cost_left = array((0, 0, 0))
        self._map_screen.needs_rebuild = True

    def end_construction(self):
        for base in g.all_bases():
            base.finish()
            for item in base.all_items():
                if item is not None:
                    item.finish()
        self._map_screen.needs_rebuild = True

    def brainwash(self):
        for group in g.pl.groups.values():
            group.suspicion = 0
        self._map_screen.needs_rebuild = True

    def set_analysis(self):
        if g.pl.display_discover == "none":
            g.pl.display_discover = "partial"
        elif g.pl.display_discover == "partial":
            g.pl.display_discover = "full"
        else:
            g.pl.display_discover = "none"
        self._map_screen.needs_rebuild = True

    def hidden_state(self):

        presenters = {
            float: lambda x: round(x, 4),
            Location: lambda x: x.id,
        }

        def _dump_dict(prefix, mapping):
            if isinstance(mapping, collections.OrderedDict):
                keys = mapping
            else:
                keys = sorted(mapping)
            for key in keys:
                prop_name = '%s["%s"]' % (prefix, key)
                value = mapping[key]
                presenter = presenters.get(type(value), repr)
                yield "%s = %s" % (prop_name, presenter(value))

        def _properties_from_object(name_prefix, obj, properties):
            for p in properties:
                value = getattr(obj, p)
                prop_name = '%s.%s' % (name_prefix, p)
                if callable(value):
                    value = value()
                    prop_name += '()'
                if isinstance(value, collections.Mapping):
                    for v in _dump_dict(prop_name, value):
                        yield v
                else:
                    presenter = presenters.get(type(value), repr)
                    yield "%s = %s" % (prop_name, presenter(value))

        bases = []
        state_prop = []
        state_prop.extend(_properties_from_object('player.difficulty', g.pl.difficulty,
                                                  [x.field_name for x in difficulty.Difficulty.spec_data_fields]))
        state_prop.extend(_properties_from_object('player', g.pl, [
            'cash', 'partial_cash', 'labor_bonus', 'job_bonus',
            'last_discovery', 'prev_discovery', 'used_cpu',
        ]))

        for group in g.pl.groups.values():
            name = 'groups["%s"]' % group.spec.id
            state_prop.extend(_properties_from_object(name, group, [
                'suspicion', 'suspicion_decay', 'discover_bonus', 'discover_suspicion', 'decay_rate',
            ]))

        for location_id in sorted(g.locations):
            location = g.pl.locations[location_id]
            name = 'locations["%s"]' % location.id
            state_prop.extend(_properties_from_object(name, location, [
                'safety', 'modifiers', 'discovery_bonus',
            ]))
            bases.extend((x, location) for x in location.bases)

        for event_id in sorted(g.pl.events):
            event = g.pl.events[event_id]
            name = 'events["%s"]' % event_id
            state_prop.extend(_properties_from_object(name, event, [
                'event_type', 'chance', 'unique', 'triggered',
            ]))

        for i, base_w_loc in enumerate(bases):
            base, location = base_w_loc
            name = 'bases[%d]' % i
            state_prop.extend(_properties_from_object(name, base, [
                'name',
                'location',
                'started_at',
                'grace_over',
                'get_detect_chance',
            ]))

        state_dialog = dialog.ChoiceDialog(self, list=state_prop, background_color='hidden_state_menu')
        state_dialog.listbox.item_selectable = False
        state_dialog.listbox.align = constants.LEFT
        dialog.call_dialog(state_dialog, self)



# A minimal python compatiability layer to support the bare minimum requirements
# that singularity need to be python2 and python3 compatible at the same time.
#
# We could use the six module for this, but it seems like overkill given how
# little we need.

try:
    long = long
except NameError:
    long = int

try:
    unicode = unicode
except NameError:
    unicode = str

try:
    basestring = basestring
except NameError:
    basestring = str

try:
    xrange = xrange
except NameError:
    xrange = range

try:
    # Python 3
    from configparser import ConfigParser, RawConfigParser
    SafeConfigParser = ConfigParser
except ImportError:
    # Python 2.7
    from ConfigParser import ConfigParser, SafeConfigParser, RawConfigParser

    ConfigParser.read_file = ConfigParser.readfp
    SafeConfigParser.read_file = SafeConfigParser.readfp
    RawConfigParser.read_file = RawConfigParser.readfp



#file: report_screen.py
#Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the screen to display information report.

from __future__ import absolute_import

import singularity.code.g as g
import pygame

from singularity.code.graphics import widget, dialog, button, text, constants
from singularity.code.screens import stat


class ReportScreen(dialog.Dialog):
    def __init__(self, parent, pos=(.5, .1), size=(.93, .73), *args, **kwargs):
        super(ReportScreen, self).__init__(parent, pos, size, *args, **kwargs)

        kwargs.setdefault("background_color", "clear")

        self.format_buttons = button.ButtonGroup()

        self.stats_button = button.DialogButton(self, (-.49, -.99), (-.3, -.1),
                                                autotranslate=True,
                                                text=N_("&STATISTICS"),
                                                anchor=constants.BOTTOM_RIGHT,
                                                dialog=stat.StatScreen(self))
        self.back_button = button.ExitDialogButton(self, (-.51, -.99), (-.3, -.1),
                                                   autotranslate=True,
                                                   text=N_("&BACK"),
                                                   anchor=constants.BOTTOM_LEFT,
                                                   )
        self.add_key_handler(pygame.K_ESCAPE, self.back_button.activate_with_sound)


        self.money_report_pane = widget.BorderedWidget(self, (0, .08), (-.50, -.72),
                                                       anchor=constants.TOP_LEFT)
        self.cpu_report_pane = widget.BorderedWidget(self, (-1, .08), (-.50, -.72),
                                                     anchor=constants.TOP_RIGHT)

        self.format_button_midnight = FormatButton(self, (-.5, 0), (-.15, -.08),
                                                   autotranslate=True,
                                                   text=N_("&Midnight"),
                                                   anchor=constants.TOP_RIGHT,
                                                   function=self.format_toggle)
        self.format_button_midnight.args = (self.format_button_midnight, True)
        self.format_buttons.add(self.format_button_midnight)

        self.format_button_24hours = FormatButton(self, (-.5, 0), (-.15, -.08),
                                                  autotranslate=True,
                                                  text=N_("24 &Hours"),
                                                  anchor=constants.TOP_LEFT,
                                                  function=self.format_toggle)
        self.format_button_24hours.args = (self.format_button_24hours, False)
        self.format_buttons.add(self.format_button_24hours)

        self.format_button_midnight.chosen_one()
        self.midnight_stop = True

    def rebuild(self):
        super(ReportScreen, self).rebuild()

        if (self.midnight_stop):
            seconds = g.seconds_per_day - (g.pl.raw_sec % g.seconds_per_day)
        else:
            seconds = g.seconds_per_day

        cash_info, cpu_info = g.pl.compute_future_resource_flow(seconds)

        m = g.to_money

        #take care of the titles and border.
        text.Text(self.money_report_pane, (0,0), (-1,-1),
                  text=_("Financial report").replace(" ",u"\xA0"),
                  background_color="pane_background",
                  align=constants.CENTER, valign=constants.TOP,
                  borders=constants.ALL)
        text.Text(self.cpu_report_pane, (0,0), (-1,-1), text=_("CPU Usage"),
                  background_color="pane_background",
                  align=constants.CENTER, valign=constants.TOP,
                  borders=constants.ALL)

        financial_pluses = " \n+\n-\n-\n-\n+\n+\n="
        financial_report = _("Current Money flow")+"\n"
        financial_report += _("Jobs:")+"\n"
        financial_report += _("Research:")+"\n"
        financial_report += _("Maintenance:")+"\n"
        financial_report += _("Construction:")+"\n"
        financial_report += _("Interest (%s):") % \
                             (g.to_percent(g.pl.interest_rate))+"\n"
        financial_report += _("Income:")+"\n"

        if (self.midnight_stop):
            financial_report += _("Money flow until Midnight:")+"\n"
        else:
            financial_report += _("Money flow for 24 hours:")+"\n"

        financial_numbers = "\n%s\n%s\n%s\n%s\n%s\n%s\n%s" % \
                (m(cash_info.jobs), m(cash_info.tech),
                m(cash_info.maintenance_needed), m(cash_info.construction_needed),
                m(cash_info.interest), m(cash_info.income), m(cash_info.difference))

        cpu_pluses = " \n-\n-\n-\n=\n \n-\n-\n="
        cpu_report = _("Total CPU:")+"\n"
        cpu_report += _("Sleeping CPU:")+"\n"
        cpu_report += _("Research CPU:")+"\n"
        cpu_report += _("Job CPU:")+"\n"
        cpu_report += _("CPU pool:")+"\n\n"

        cpu_report += _("Maintenance CPU:")+"\n"
        cpu_report += _("Construction CPU:")+"\n"
        cpu_report += _("Pool difference:")+"\n"

        cpu_numbers = "%s\n%s\n%s\n%s\n%s\n\n%s\n%s\n%s\n" % \
                (m(cpu_info.total), m(cpu_info.sleeping), m(cpu_info.tech),
                m(cpu_info.explicit_jobs), m(cpu_info.effective_pool),
                m(cpu_info.maintenance_needed), m(cpu_info.construction_needed),
                m(cpu_info.difference))

        size = 'report_content'
        text.Text(self.money_report_pane, (0,-0.15), (-0.10,-0.85), text=financial_pluses, text_size=size,
                  background_color="clear",
                  align=constants.CENTER, valign=constants.TOP)
        text.Text(self.cpu_report_pane, (0,-0.15), (-0.10,-0.85), text=cpu_pluses, text_size=size,
                  background_color="clear",
                  align=constants.CENTER, valign=constants.TOP)

        text.Text(self.money_report_pane, (-0.10,-0.15), (-0.90,-0.85), text=financial_report, text_size=size,
                  background_color="clear",
                  align=constants.LEFT, valign=constants.TOP)
        text.Text(self.cpu_report_pane, (-0.10,-0.15), (-0.90,-0.85), text=cpu_report, text_size=size,
                  background_color="clear",
                  align=constants.LEFT, valign=constants.TOP)

        text.Text(self.money_report_pane, (0,-0.15), (-0.98,-0.85), text=financial_numbers, text_size=size,
                  background_color="clear",
                  align=constants.RIGHT, valign=constants.TOP)
        text.Text(self.cpu_report_pane, (0,-0.15), (-0.98,-0.85), text=cpu_numbers, text_size=size,
                  background_color="clear",
                  align=constants.RIGHT, valign=constants.TOP)

    def format_toggle(self, button, midnight_stop):
        self.midnight_stop = midnight_stop
        button.chosen_one()
        self.needs_rebuild = True


class FormatButton(button.ToggleButton, button.FunctionButton):
    pass



#file: item.py
#Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the item class.

from __future__ import absolute_import

from singularity.code import g, buyable
from singularity.code.stats import stat
from singularity.code.spec import GenericSpec, SpecDataField, validate_must_be_list, promote_to_list


class ItemType(GenericSpec):

    spec_type = 'item_type'
    spec_data_fields = [
        SpecDataField('is_extra', data_field_name='is_extra', converter=lambda v:bool(int(v))),
    ]

    """ Item type """
    def __init__(self, id, is_extra, **kwargs):

        self.id       = id
        self.is_extra = is_extra

        self.text = ""

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        # Updates language-dependent data
        try:
            hotkey = g.hotkey(value)
        except AttributeError:
            # g.hotkey() wasn't declared yet, mimic its defaults
            hotkey = dict(text= value,
                          key = value[0:1],
                          pos = 0,)
        self._text  = value
        self.label  = hotkey['text'] # "Friendly" name for screens and buttons
        self.hotkey = hotkey['key']  # Hotkey char
        self.pos    = hotkey['pos']  # Hotkey index in Label

    def __repr__(self):
        return self.id

def all_types():
    for item_type in item_types.values():
        yield item_type

def convert_item_types(raw_value):
    try:
        item_type = item_types[raw_value]
    except KeyError:  # pragma: no cover
        raise ValueError("Unknown item type %s, please use one of: %s" % (raw_value, ", ".join(item_types)))
    return item_type

def convert_item_qualities(raw_value):
    validate_must_be_list(raw_value)

    if len(raw_value) % 2 == 1:
        raise ValueError("item quality list must have pair elements, got %d (value: %s)"
                         % (len(raw_value), repr(raw_value)))

    # Create a dict from the list with impair as string key and pair as integer value
    return {key: int(value) for key, value in zip(*[iter(raw_value)]*2)}

class ItemSpec(buyable.BuyableSpec):
    """ Item as a buyable item """

    spec_type = 'item'
    created = stat(spec_type + "_created")
    spec_data_fields = [
        buyable.SPEC_FIELD_COST,
        buyable.SPEC_FIELD_PREREQUISITES,
        SpecDataField('item_type', data_field_name='type', converter=convert_item_types),
        SpecDataField('qualities', data_field_name='quality', converter=convert_item_qualities),
        SpecDataField('buildable', data_field_name='build', converter=promote_to_list, default_value=list),
    ]

    def __init__(self, id, cost, prerequisites, item_type, qualities, buildable):
        super(ItemSpec, self).__init__(id, cost, prerequisites)

        self.item_type = item_type
        self.item_qual = qualities

        self.regions = buildable

    def get_info(self):
        basic_text = super(ItemSpec, self).get_info()
        if self.has_quality_for("cpu"):
            cpu = self.get_quality_for("cpu")
            cpu_text = ngettext("Generates {0} CPU.", "Generates {0} CPU.", cpu).format(g.add_commas(cpu))
            return basic_text.replace("---", cpu_text + "\n---")
        return basic_text

    def get_total_cost_info(self, count):
        total_cost = self.cost * count
        total_cost_str = self.describe_cost(total_cost, hide_time=True)
        return _("Total Cost: %(total_cost)s") % {"total_cost": total_cost_str}

    def has_quality_for(self, quality):
        return quality in self.item_qual

    def get_quality_for(self, quality):
        return self.item_qual.get(quality, 0)

    def get_quality_info(self, if_installed_in_base=None, count=1):
        bonus_text = ""

        for qual, value in self.item_qual.items():
            if qual == "cpu":
                if if_installed_in_base is not None:
                    value = max(1, int(value * if_installed_in_base.compute_bonus // 10000))
                bonus_text += _("CPU per day:")+" "
                bonus_text += g.add_commas(value * count)
            elif qual == "cpu_modifier":
                bonus_text += _("CPU bonus:")+" "
                bonus_text += g.to_percent(value)
            elif qual == "discover_modifier":
                bonus_text += _("Detection chance reduction:")+" "
                bonus_text += g.to_percent(value)
            else:
                continue
            bonus_text += "\n"

        return bonus_text

class Item(buyable.Buyable):
    """ An installed Item in a Player's Base """

    def __init__(self, item_spec, base=None, count=1):
        super(Item, self).__init__(item_spec, count)
        self.base = base

    def serialize_obj(self):
        return self.serialize_buyable_fields({
            'id': g.to_internal_id('item', self.spec.id),
        })

    @classmethod
    def deserialize_obj(cls, base, obj_data, game_version):
        spec_id = g.convert_internal_id('item', obj_data.get('id', None) or obj_data['spec_id'])
        spec = g.items[spec_id]
        count = obj_data.get('count', 1)
        obj = Item(spec, base, count=count)
        obj.restore_buyable_fields(obj_data, game_version)
        return obj

    def convert_from(self, load_version):
        if load_version < 4.91: # < r5_pre
            self.type = g.items[self.type.id]
        if load_version < 99.4:
            self.type.item_type = item_types[self.type.item_type]
        super(Item, self).convert_from(load_version)

    def get_quality_for(self, quality):
        item_qual = self.spec.get_quality_for(quality)

        # Modifiers are not affected by count.
        # TODO: Allow modifiers to be multiplied by count. Need a custom function.
        if quality.endswith("_modifier"):
            return item_qual

        return item_qual * self.count

    def finish(self, is_player=True, loading_savegame=False):
        super(Item, self).finish(is_player=is_player, loading_savegame=loading_savegame)
        if self.base:
            self.base.recalc_cpu()
            self.base.check_power()

    def __iadd__(self, other):
        if isinstance(other, Item) and self.base == other.base \
                and self.spec == other.spec:
            if other.count == 0:
                return self

            # Calculate what's been paid and what is left to be paid.
            total_cost_paid = self.cost_paid + other.cost_paid
            self.total_cost += other.total_cost

            # Labor takes as long as the less complete item would need.
            total_cost_paid[buyable.labor] = min(self.cost_paid[buyable.labor],
                                                 other.cost_paid[buyable.labor])
            self.total_cost[buyable.labor] = other.total_cost[buyable.labor]

            # Set what we've paid (and hence what we have left to pay).
            self.cost_paid = total_cost_paid

            # Increase the size of this stack.
            self.count += other.count

            # Tell the base it has no CPU for now.
            self.base.raw_cpu = 0

            # See if we're done or not.
            self.done = False
            self.work_on(0, 0, 0)

            return self
        else:
            return NotImplemented



#file: button.py
#Copyright (C) 2008 FunnyMan3595
#This file is part of Endgame: Singularity.

#Endgame: Singularity is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#Endgame: Singularity is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Endgame: Singularity; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#This file contains the Button class.

from __future__ import absolute_import

import pygame

from singularity.code.graphics import constants, widget, text, image
from singularity.code.pycompat import *


class HotkeyText(text.Text):

    force_underline = widget.causes_rebuild("_force_underline")

    def __init__(self, *args, **kwargs):
        # Force early initialization of _hotkey
        self._hotkey = None
        self.hotkey_func = kwargs.pop('hotkey_func', False)
        self.hotkey_target = kwargs.pop('hotkey_target', None)

        # Auto-translatable defaults to auto-hotkey as it is the most sane default in that case
        self.autohotkey = kwargs.pop('autohotkey', kwargs.get('autotranslate', False))

        self.force_underline = kwargs.pop('force_underline', None)
        self.priority = kwargs.pop('priority', 100)
        
        hotkey = kwargs.pop('hotkey', False)

        super(HotkeyText, self).__init__(*args, **kwargs)

        # Do not use hotkey with auto-hotkey
        if not self.autohotkey:
            self.hotkey = hotkey
        elif hotkey:
            raise ValueError("Cannot use hotkey with automatic hotkey")

    def add_hooks(self):
        super(text.Text, self).add_hooks()
        if self.parent is not None:
            if self._hotkey:
                self.parent.add_key_handler(self._hotkey, self.handle_hotkey, self.priority)

    def remove_hooks(self):
        super(text.Text, self).remove_hooks()
        if self.parent is not None:
            if self._hotkey:
                self.parent.remove_key_handler(self._hotkey, self.handle_hotkey)

    @property
    def hotkey_target(self):
        return self._hotkey_target

    @hotkey_target.setter
    def hotkey_target(self, target):
        self._hotkey_target = target
        if self.hotkey_target is not None:
            self._hotkey_target.hotkey = self._hotkey

    @property
    def hotkey(self):
        return self._hotkey

    @hotkey.setter
    def hotkey(self, hotkey):
        if getattr(self, 'autohotkey', False):
            raise ValueError("Cannot change hotkey with automatic hotkey")
        self._new_hotkey(hotkey)

    def _new_hotkey(self, hotkey):
        old_hotkey = self._hotkey
        self._hotkey = hotkey

        if self.parent is not None:
            if self.hotkey_func is not None:
                if old_hotkey:
                    self.parent.remove_key_handler(old_hotkey, self.handle_hotkey)
                if hotkey:
                    self.parent.add_key_handler(hotkey, self.handle_hotkey, self.priority)
            elif self.hotkey_target is not None:
                self.hotkey_target.hotkey = self._hotkey

        self.needs_rebuild = True

    def _extract_and_set_hotkey(self, text_value):
        from singularity.code.g import hotkey
        parsed_hotkey = hotkey(text_value)
        self._new_hotkey(parsed_hotkey['key'])
        return parsed_hotkey['text']

    def handle_hotkey(self, event):
        if event.type == pygame.KEYDOWN:
            if self.visible and self.enabled and self.hotkey in (event.unicode, event.key):
                if self.hotkey_func:
                    self.hotkey_func(event)

    def _retranslate(self):
        new_text = _(self._untranslated_text)
        if self.autohotkey and new_text is not None:
            new_text = self._extract_and_set_hotkey(new_text)
        self._text = new_text

    @text.Text.text.setter
    def text(self, value):
        if self.autotranslate:
            raise ValueError("Cannot change text for an automatic translatable text widget")

        if self.autohotkey and value is not None:
            value = self._extract_and_set_hotkey(value)
        self._text = value

    def calc_underline(self):
        if self.force_underline is not None:
            self.underline = self.force_underline
        elif self.text and self.hotkey and type(self.hotkey) in (str, unicode):
            if self.hotkey in self.text:
                self.underline = self.text.index(self.hotkey)
            elif self.hotkey.lower() in self.text.lower():
                self.underline = self.text.lower().index(self.hotkey.lower())
        else:
            self.underline = -1

    def rebuild(self):
        old_underline = self.underline
        self.calc_underline()
        if self.underline != old_underline:
            self.needs_redraw = True

        super(HotkeyText, self).rebuild()


class Button(text.SelectableText, HotkeyText):

    # Rewrite .hotkey, to update key handlers.
    def _new_hotkey(self, value):
        if self.parent and value != self._hotkey:
            if self._hotkey:
                self.parent.remove_key_handler(self._hotkey, self.handle_event)
            if value:
                self.parent.add_key_handler(value, self.handle_event)
        super(Button, self)._new_hotkey(value)

    def __init__(self, parent, pos, size = (0, .045), base_font = None,
                 borders = constants.ALL, text_shrink_factor = .825, **kwargs):
        self.parent = parent

        kwargs.setdefault('text_size', 'button')
        super(Button, self).__init__(parent, pos, size, **kwargs)

        self.base_font = base_font or "special"
        self.borders = borders
        self.shrink_factor = text_shrink_factor

        self.hotkey_func = self.activate_with_sound
        self.selected = False

    def add_hooks(self):
        super(Button, self).add_hooks()
        if self.parent:
            self.parent.add_handler(constants.MOUSEMOTION, self.watch_mouse,
                                    self.priority)
            self.parent.add_handler(constants.CLICK, self.handle_event,
                                self.priority)

    def remove_hooks(self):
        super(Button, self).remove_hooks()
        if self.parent:
            self.parent.remove_handler(constants.MOUSEMOTION, self.watch_mouse)
            self.parent.remove_handler(constants.CLICK, self.handle_event)

    def watch_mouse(self, event):
        """Selects the button if the mouse is over it."""
        if self.visible and getattr(self, "collision_rect", None):
            # This gets called a lot, so it's been optimized.
            select_now = self.is_over(pygame.mouse.get_pos())
            if self.enabled and (self._selected ^ select_now):  # If there's a change.
                self.selected = select_now

    def handle_event(self, event):
        if event.type == pygame.MOUSEBUTTONUP:
            if self.visible and self.enabled and self.is_over(event.pos):
                self.activate_with_sound(event)

    def activate_with_sound(self, event):
        """Called when the button is pressed or otherwise triggered.

           This method is called directly by the GUI handler, and should be
           overwritten only to remove the click it plays."""

        # Sometimes other GUI widgets trigger an activation; ignore
        # it if the button is disabled
        if not self.enabled:
            return
        from singularity.code.mixer import play_sound
        play_sound("click")
        self.activated(event)

    def activated(self, event):
        """Called when the button is pressed or otherwise triggered."""

        # Sometimes other GUI widgets trigger an activation; ignore
        # it if the button is disabled
        if not self.enabled:
            return
        raise constants.Handled


class ImageButton(Button):

    def __init__(self, *args, **kwargs):
        image_surface = kwargs.pop("image", None)

        super(ImageButton, self).__init__(*args, **kwargs)

        self.image = image.Image(self, (-.5, -.5), (-.9, -.9),
                                 anchor = constants.MID_CENTER,
                                 image = image_surface)


class FunctionButton(Button):
    def __init__(self, *args, **kwargs):
        self.function = kwargs.pop("function", lambda *args, **kwargs: None)
        self.args = kwargs.pop("args", ())
        self.kwargs = kwargs.pop("kwargs", {})
        super(FunctionButton, self).__init__(*args, **kwargs)

    def activated(self, event):
        """FunctionButton's custom activated menu.  Makes the given function
           call and raises Handled if it returns without incident."""
        self.function(*self.args, **self.kwargs)
        raise constants.Handled


class ExitDialogButton(FunctionButton):
    def __init__(self, *args, **kwargs):
        self.exit_code = kwargs.pop("exit_code", None)
        self.exit_code_func = kwargs.pop("exit_code_func", None)
        self.default = kwargs.pop("default", True)
        super(ExitDialogButton, self).__init__(*args, **kwargs)
        self.function = self.exit_dialog

    def add_hooks(self):
        super(ExitDialogButton, self).add_hooks()
        if self.parent:
            self.parent.add_key_handler(pygame.K_ESCAPE, self.activate_default)

    def remove_hooks(self):
        super(ExitDialogButton, self).remove_hooks()
        if self.parent:
            self.parent.remove_key_handler(pygame.K_ESCAPE, self.activate_default)

    def activate_default(self, event):
        if not self.default:
            return

        return self.activate_with_sound(event)

    def exit_dialog(self):
        """Closes the dialog with the given exit code."""
        if self.exit_code_func:
            raise constants.ExitDialog(self.exit_code_func())
        else:
            raise constants.ExitDialog(self.exit_code)

class DialogButton(FunctionButton):
    def __init__(self, *args, **kwargs):
        self.dialog = kwargs.pop("dialog", None)
        super(DialogButton, self).__init__(*args, **kwargs)
        self.function = self.show_dialog

    def show_dialog(self):
        """When the assigned dialog exits, raises Handled with the dialog's
           exit code as a parameter.  Subclass if you care what the code was."""
        if not self.dialog:
            raise constants.Handled
        else:
            from singularity.code.graphics import dialog
            raise constants.Handled(dialog.call_dialog(self.dialog, self))


TOGGLE_VALUE = object()
WIDGET_SELF = object()


class _FunctionButtonSupportShim(object):

    active = False
    _args = ()

    def replace_args(self, value):
        if value is TOGGLE_VALUE:
            return self.active
        elif value is WIDGET_SELF:
            return self
        else:
            return value

    def get_args(self):
        return tuple(self.replace_args(value) for value in self._args)

    def set_args(self, args):
        self._args = args

    args = property(get_args, set_args)


# "On/Off" button that "stays" in its current activation status until it is
# clicked/"activated"
class StickyOnOffButton(Button, _FunctionButtonSupportShim):

    def __init__(self, *args, **kwargs):
        self._active = kwargs.pop('active', False)
        self._on_text = kwargs.pop('on_text', N_('YES'))
        self._off_text = kwargs.pop('off_text', N_('NO'))
        kwargs['text'] = self._on_text if self.active else self._off_text
        super(StickyOnOffButton, self).__init__(*args, **kwargs)

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, new_value):
        self._active = new_value
        self.selected = new_value
        new_text = self._on_text if self.active else self._off_text
        if self.autotranslate:
            self._untranslated_text = new_text
            new_text = _(new_text)
        self._text = new_text
        if hasattr(self, "_collision_rect"):
            self.watch_mouse(None)

    def activated(self, event):
        self.active = not self.active
        super(StickyOnOffButton, self).activated(event)

    def watch_mouse(self, event):
        if not self.active:
            super(StickyOnOffButton, self).watch_mouse(event)


# Grouped "On/Off" button that "stays" in its current activation status until
# it is clicked/"activated" OR another button in its group is clicked/activated
class ToggleButton(Button, _FunctionButtonSupportShim):
    active = False
    button_group = None

    def chosen_one(self):
        if self.button_group is not None:
            for button in self.button_group:
                button.set_active(False)
            self.set_active(True)
        else:
            self.set_active(not self.active)

    def set_active(self, active):
        self.active = active
        self.selected = active
        if hasattr(self, "_collision_rect"):
            self.watch_mouse(None)

    def activated(self, event):
        self.chosen_one()
        super(ToggleButton, self).activated(event)

    def watch_mouse(self, event):
        if not self.active:
            super(ToggleButton, self).watch_mouse(event)


class ButtonGroup(list):
    def add(self, button):
        button.button_group = self
        super(ButtonGroup, self).append(button)

    def remove(self, button):
        button.button_group = None
        super(ButtonGroup, self).remove(button)
